# import context
import argparse
import os
from termcolor import colored
from ..model.common.modelloader import ExcelMaker, valid_programs, XmlReader


def generate_file(program_code, excel_file):
    em = ExcelMaker(program_code)
    em.makeExcelBasedOnModel(excel_file)
    print(colored(f"{excel_file} is created...", "green"))


def main():
    parser = argparse.ArgumentParser(
        description="used for processing temporary resident visa, study permit, and work permit both in Canada or outside of Canada"
    )
    parser.add_argument(
        "-p",
        "--program",
        help="input program code. 5257 for trv, 1294 or 1295 for sp or wp outside of Canada. 5708/5709/5710 for vr/sp/wp in Canada ",
    )
    parser.add_argument("-x", "--xml", help="input xml file name")
    parser.add_argument("-t", "--to", help="input excel file name for output")
    args = parser.parse_args()

    if args.program not in valid_programs:
        print(colored(f"Your input program code {args.program} is not valid. ", "red"))
        return

    if args.to and not os.path.isfile(args.to):
        generate_file(args.program, args.to)
        return

    if args.program and args.xml and args.to:
        xr = XmlReader(args.program)
        xr.readXmlData2Excel(args.xml, args.to)
        print(colored(f"Excel file {args.to} is filled with {args.xml} data", "green"))
        return

    print(colored("Invalid command and flag combination... try trv -h for help", "red"))


if __name__ == "__main__":
    main()



from typing import List
from model.common.commonmodel import CommonModel
from model.bcpnp.data import Bcpnp, Contact, General, JobOffer, ErAddress
from model.common.jobposition import PositionBase
from model.common.rcic import Rcic
from model.common.advertisement import (
    Advertisement,
    Advertisements,
    InterviewRecord,
    InterviewRecords,
    RecruitmentSummary,
)
from model.common.person import Person, PersonalAssess
from model.common.contact import Contacts
from model.common.address import Addresses
from model.common.phone import Phone, Phones
from datetime import date
from model.common.xmlfiller import XmlFiller
import os


class Personal(Person):
    def __str__(self):
        return self.full_name


class Position(PositionBase):
    pass


class JobOfferFormModel(CommonModel):
    eraddress: List[ErAddress]
    phone: List[Phone]
    general: General
    contact: List[Contact]
    position: Position
    personal: Personal
    joboffer: JobOffer
    personalassess: PersonalAssess
    bcpnp: Bcpnp
    rcic: Rcic
    advertisement: List[Advertisement]
    interviewrecord: List[InterviewRecord]
    recruitmentsummary: RecruitmentSummary

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        if output_excel_file:
            excels = self.getExcels(
                [
                    "excel/er.xlsx",
                    "excel/pa.xlsx",
                    "excel/recruitment.xlsx",
                    "excel/bcpnp.xlsx",
                    "excel/rep.xlsx",
                ]
            )
        else:
            if excels is None and len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        # call parent class for validating
        super().__init__(excels, output_excel_file, globals())

    @property
    def work_location(self):
        addresses = Addresses(self.eraddress)
        return addresses.working

    @property
    def phones(self):
        return Phones(self.phone)

    @property
    def selected_contact(self):
        contacts = Contacts(self.contact)
        return contacts.preferredContact

    @property
    def interviews(self):
        return InterviewRecords(self.interviewrecord)

    @property
    def advertisements(self):
        return Advertisements(self.advertisement)

    @property
    def businessaddress(self):
        eraddress = Addresses(self.eraddress)
        return eraddress.business

    @property
    def mailingaddress(self):
        eraddress = Addresses(self.eraddress)
        return eraddress.mailing

    @property
    def person(self):
        return {
            "first_name": self.personal.first_name,
            "last_name": self.personal.last_name,
            "full_name": self.personal.full_name,
            "attributive": self.personal.attributive,
            "object": self.personal.object,
            "subject": self.personal.subject,
            "short_name": self.personal.short_name,
            "why_tfw": self.personalassess.why_qualified_say,
        }


class JobOfferFormDocxAdaptor:
    def __init__(self, jobofferform_obj: JobOfferFormModel):
        self.jobofferform_obj = jobofferform_obj

    def re_generate_dict(self):
        summary_info = {
            "personal": self.jobofferform_obj.person,
            "date_of_offer": self.jobofferform_obj.joboffer.date_of_offer,
            "work_start_date": self.jobofferform_obj.joboffer.start_date_say,
            "joboffer_date": self.jobofferform_obj.joboffer.date_of_offer,
            "annual_rate": float(self.jobofferform_obj.joboffer.annual_rate),
            "hourly_rate": float(self.jobofferform_obj.joboffer.hourly_rate),
            "work_location": self.jobofferform_obj.work_location,
            "b_address": self.jobofferform_obj.businessaddress.line1,
            "b_city": self.jobofferform_obj.businessaddress.city,
            "b_province": self.jobofferform_obj.businessaddress.province,
            "b_country": self.jobofferform_obj.businessaddress.country,
            "b_postcode": self.jobofferform_obj.businessaddress.post_code,
            "m_address": self.jobofferform_obj.mailingaddress.line1,
            "m_city": self.jobofferform_obj.mailingaddress.city,
            "m_province": self.jobofferform_obj.mailingaddress.province,
            "m_country": self.jobofferform_obj.mailingaddress.country,
            "m_postcode": self.jobofferform_obj.mailingaddress.post_code,
            "w_address": self.jobofferform_obj.work_location.line1,
            "w_city": self.jobofferform_obj.work_location.city,
            "w_province": self.jobofferform_obj.work_location.province,
            "w_country": self.jobofferform_obj.work_location.country,
            "w_postcode": self.jobofferform_obj.work_location.post_code,
            "date_signed": date.today(),
            # position 3c
            # TODO: is new position , ...
            # "under_cba":
            "which_union": self.jobofferform_obj.position.which_union,
            "lmia_refused": "Yes_5"
            if self.jobofferform_obj.position.lmia_refused
            else "No_5",
            "lmia_refused_reason": self.jobofferform_obj.position.lmia_refused_reason
            if self.jobofferform_obj.position.lmia_refused
            else "",
            # "other_language_required":self.jobofferform_obj.joboffer.other_language_required #TODO: ,
            "what_language": self.jobofferform_obj.joboffer.reason_for_other,
            # "license_required":
            "which_license": self.jobofferform_obj.joboffer.license_description,
            # 4. recruitment summary
            "active_recruitment": "Yes_7"
            if self.jobofferform_obj.advertisements.amount > 0
            else "No_7",
            "min_days": self.jobofferform_obj.advertisements.min_days,
            "total_applicant": self.jobofferform_obj.interviews.resume_num,
            "recruitment_summary": self.jobofferform_obj.advertisements.summary
            + self.jobofferform_obj.interviews.summary
            + self.jobofferform_obj.position.how_did_you_find,
            "contact_email": self.jobofferform_obj.selected_contact.email,
            "contact_last_name": self.jobofferform_obj.selected_contact.last_name,
            "contact_first_name": self.jobofferform_obj.selected_contact.first_name,
            "contact_phone": self.jobofferform_obj.selected_contact.phone,
            "contact_position": self.jobofferform_obj.selected_contact.position,
            "business_phone": self.jobofferform_obj.phones.business.international_format_full,
        }
        return {**self.jobofferform_obj.dict(), **summary_info}

    def make(self, output_xml):
        path = os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
        template_path = path + "/template/xml/bcpnp_job_offer_form.xml"
        xf = XmlFiller(template_path, self.re_generate_dict())
        xf.save(output_xml)



from pydantic import BaseModel, validator, root_validator
from typing import Optional, List
from termcolor import colored
import re


class Address(BaseModel):
    variable_type: str  # this is the identifier variable
    display_type: str  # actually it is the prompt information
    po_box: Optional[str]
    unit: Optional[str]
    street_number: Optional[str]
    street_name: Optional[str]
    district: Optional[str]
    city: Optional[str]
    province: Optional[str]
    country: Optional[str]
    post_code: Optional[str]

    @validator("*", pre=True)
    def clean_str(cls, v):
        if isinstance(v, str):
            return re.sub(r"\s+", " ", v.strip())
        return v

    @root_validator
    def checkCanadaProvince(cls, values):
        country = values["country"]
        province = values["province"]
        street_name = values["street_name"]
        # binding with street number is for those like work location: Canada is always selected by actually no data in some rows
        if (
            country
            and country.lower() == "canada"
            and street_name
            and province
            not in [
                "AB",
                "BC",
                "MB",
                "NB",
                "NL",
                "NS",
                "NT",
                "NU",
                "ON",
                "PE",
                "QC",
                "SK",
                "YT",
            ]
        ):
            raise ValueError(
                colored(
                    f'Since country is Canada is, the province must be one of  "AB","BC","MB","NB","NL","NS","NT","NU","ON","PE","QC","SK","YT"',
                    "red",
                )
            )
        return values

    @root_validator
    def checkRowCompletion(cls, values):
        all_fields = [
            "po_box",
            "unit",
            "street_number",
            "street_name",
            "city",
            "district",
            "province",
            "country",
            "post_code",
        ]
        all_fields_values = [values[field] for field in all_fields]

        required_fields = [
            "street_number",
            "street_name",
            "city",
            "country",
            "post_code",
        ]
        required_values = [values[field] for field in required_fields]

        has_values = [value for value in required_values if value]

        if any(all_fields_values) and not all(required_values):
            raise ValueError(
                f"Please check the row with values ({','.join(has_values)}), some required fileds are missed."
            )
        return values

    @property
    def full_address(self):
        fa = self.po_box + " " if self.po_box else ""
        fa += self.unit + " " if self.unit else ""
        fa += self.street_number + " " if self.street_number else ""
        fa += self.street_name + ", "
        fa += self.district + ", " if self.district else ""
        fa += self.city + ", "
        fa += self.province + ", " if self.province else ""
        fa += self.country + " " + self.post_code

        return fa

    @property
    def line1(self):
        l1 = self.po_box + " " if self.po_box else ""
        l1 += self.unit + " " if self.unit else ""
        l1 += self.street_number + " " if self.street_number else ""
        l1 += self.street_name
        l1 += ", " + self.district + ", " if self.district else ""
        return l1

    @property
    def line2(self):
        l2 = (
            self.city
            + ", "
            + self.province
            + ", "
            + self.country
            + " "
            + self.post_code
        )
        return l2

    @property
    def street_address(self):
        l1 = self.street_number + " " if self.street_number else ""
        l1 += self.street_name if self.street_name else ""
        l1 += ", " + self.district + ", " if self.district else ""
        return l1

    def __eq__(self, another):
        for k, v in self.__dict__.items():
            if k not in ["variable_type", "display_type"] and v != getattr(
                another, k, None
            ):
                return False
        return True

    def __str__(self):
        return self.full_address


class Addresses(object):
    def __init__(self, address_list: List[Address]) -> None:
        self.addresses = address_list

    def _specific_address(self, v_type):
        address = [
            address for address in self.addresses if address.variable_type == v_type
        ]
        return address[0] if address else None

    @property
    def mailing(self):
        return self._specific_address("mailing_address")

    @property
    def residential(self):
        return self._specific_address("residential_address")

    @property
    def business(self):
        return self._specific_address("business_address")

    # return the first work location
    @property
    def working(self):
        return self._specific_address("working_address")

    # return all work locations
    @property
    def workings(self):
        return [
            address
            for address in self.addresses
            if address.variable_type == "working_address"
        ]

    @property
    def PreferredAddress(self):
        for addr_type in [
            "business_address",
            "residential_address",
            "mailing_address",
            "working_address",
        ]:
            addr = self._specific_address(addr_type)
            if addr:
                return addr

    def getPreferredAddressFromList(self, type_list):
        for addr_type in type_list:
            addr = self._specific_address(addr_type)
            if addr:
                return addr



from email.errors import ObsoleteHeaderDefect
from optparse import Option
from typing import List, Optional, Union
from model.common.address import Address
from model.common.educationbase import EducationBase
from model.common.phone import Phone
from model.common.commonmodel import CommonModel
from model.common.employmentbase import EmploymentBase
from model.common.jobofferbase import JobofferBase
from model.common.id import ID
from model.common.person import Person
from model.common.cor import CORs, COR
from model.common.status import Status
from datetime import date, datetime
from pydantic import BaseModel, EmailStr, root_validator, validator
from model.common.utils import normalize
from webform.prportal.data.country import country_residence
from model.common.contact import ContactBase
from model.common.family import FamilyBase
from termcolor import colored


class PersonId(ID):
    pass


class CanadaRelative(BaseModel):
    last_name: Optional[str]
    first_name: Optional[str]
    sex: Optional[str]
    relationship: Optional[str]
    status: Optional[str]
    age: Optional[int]
    city: Optional[str]
    province: Optional[str]
    years_in_canada: Optional[int]

    @root_validator
    def checkFields(cls, values):
        last_name = values["last_name"]
        first_name = values["first_name"]
        sex = values["sex"]
        relationship = values["relationship"]
        status = values["status"]
        age = values["age"]
        city = values["city"]
        province = values["province"]
        years_in_canada = values["years_in_canada"]

        if (
            last_name
            and first_name
            and not all(
                [sex, relationship, status, age, city, province, years_in_canada]
            )
        ):
            raise ValueError(
                f"Some fields in row of {first_name} {last_name} are missed."
            )
        return values


class Rcic(BaseModel):
    first_name: str
    last_name: str
    telephone: str


class General(BaseModel):
    legal_name: str
    operating_name: Optional[str]
    corporate_structure: str
    other_explaination: Optional[str]
    establish_date: date
    industry: str
    registration_number: str
    ft_employee_number: int
    establish_date: date
    website: Optional[str]


class Contact(ContactBase):
    position: Optional[str]

    @root_validator
    def checkFields(cls, values):
        last_name = values["last_name"]
        first_name = values["first_name"]
        phone = values["phone"]
        email = values["email"]
        position = values["position"]

        if last_name and first_name and not all([phone, email, position]):
            raise ValueError(
                f"Some fields in row of {first_name} {last_name} are missed."
            )
        return values


class ErAddress(Address):
    phone: Optional[str]


class Joboffer(JobofferBase):
    phone: str
    is_working: bool
    license_request: bool


class Employment(EmploymentBase):
    bcpnp_qualified: bool
    phone_of_certificate_provider: Optional[str]
    website: Optional[str]
    unit: Optional[str]
    street_address: Optional[str]
    city: Optional[str]
    province: Optional[str]
    country: Optional[str]
    postcode: Optional[str]
    duties: Optional[str]


class Marriage(BaseModel):
    marital_status: Optional[str]
    married_date: Optional[date]
    sp_last_name: Optional[str]
    sp_first_name: Optional[str]
    sp_in_canada: Optional[bool]
    sp_canada_status: Optional[str]
    sp_canada_status_end_date: Optional[date]
    sp_in_canada_other: Optional[str]
    sp_in_canada_work: Optional[bool]
    sp_canada_occupation: Optional[str]
    sp_canada_employer: Optional[str]

    @root_validator
    def checkCurrentMarriage(cls, values):
        marital_status = values["marital_status"]
        married_date = values["married_date"]
        sp_last_name = values["sp_last_name"]
        sp_first_name = values["sp_first_name"]

        if marital_status in ["Common-Law", "Married"] and not all(
            [married_date, sp_last_name, sp_first_name]
        ):
            raise ValueError(
                f"Marital stauts is {marital_status}, but you did not answer all the questions for spouse: first/last name, or marriage date"
            )
        return values

    @root_validator
    def checkCanadaStatus(cls, values):
        sp_in_canada = values["sp_in_canada"]
        sp_canada_status = values["sp_canada_status"]
        sp_canada_status_end_date = values["sp_canada_status_end_date"]
        sp_in_canada_other = values["sp_in_canada_other"]

        if sp_in_canada and not all([sp_canada_status, sp_canada_status_end_date]):
            raise ValueError(
                f"Spouse is in Canada, but you did not answer all the questions for spouse status and/or status end date",
                "red",
            )
        if sp_canada_status == "Other" and not sp_in_canada_other:
            raise ValueError(
                f"Spouse is in Canada and status is other, but you did not give the explanation"
            )
        return values

    @root_validator
    def checkWorkInCanada(cls, values):
        sp_in_canada_work = values["sp_in_canada_work"]
        sp_canada_occupation = values["sp_canada_occupation"]
        sp_canada_employer = values["sp_canada_employer"]

        if sp_in_canada_work and not all([sp_canada_occupation, sp_canada_employer]):
            raise ValueError(
                colored(
                    f"Spouse is working in Canada, but you did not answer all the questions for spouse occupation and/or employer",
                    "red",
                )
            )
        return values


class Personal(Person):
    uci: Optional[str]

    _normalize_used_first_name = validator(
        "used_first_name", allow_reuse=True, check_fields=False
    )(normalize)
    _normalize_used_last_name = validator(
        "used_last_name", allow_reuse=True, check_fields=False
    )(normalize)

    @property
    def user_id(self):
        dob = self.dob.strftime(("%Y-%m-%d"))
        return (
            self.last_name[0].upper()
            + self.first_name[0]
            + dob.split("-")[0]
            + dob.split("-")[1]
            + dob.split("-")[2]
        )

    @property
    def password(self):
        return "Super" + str(datetime.today().year) + "!"


class Education(EducationBase):
    city: Optional[str]
    province: Optional[str]
    country: Optional[str]
    is_trade: Optional[bool]
    graduate_date: Optional[date]

    @root_validator
    def checkCompletion(cls, values):
        school_name = values["school_name"]
        city = values["city"]
        province = values["province"]
        country = values["country"]
        is_trade = values["is_trade"]
        graduate_date = values["graduate_date"]
        education_level = values["education_level"]
        field_of_study = values["field_of_study"]

        if school_name and not all(
            [
                city,
                province,
                country,
                is_trade != None,
                graduate_date,
                education_level,
                field_of_study,
            ]
        ):
            raise ValueError(f"Some fields are missed in row of {school_name}.")
        return values


class Family(FamilyBase):
    country_of_citizenship: Optional[str]
    address: Optional[str]
    date_of_death: Optional[date]
    marital_status: Optional[str]

    @root_validator
    def checkFields(cls, values):
        last_name = values["last_name"]
        first_name = values["first_name"]
        date_of_birth = values["date_of_birth"]
        place_of_birth = values["place_of_birth"]
        birth_country = values["birth_country"]
        marital_status = values["marital_status"]
        address = values["address"]

        if (
            last_name
            and first_name
            and not all(
                [
                    date_of_birth,
                    place_of_birth,
                    birth_country,
                    marital_status,
                    address,
                ]
            )
        ):
            raise ValueError(
                f"Some fields of {first_name} {last_name} in info-family sheet are not completed, please check"
            )
        return values


class Bcpnp(BaseModel):
    intended_city: str
    account: Optional[str]
    password: Optional[str]
    # has_current_app: bool
    # has_applied_before: bool
    # pre_file_no: Optional[str]
    case_stream: Optional[str]
    q1: bool
    q1_explaination: Optional[str]
    q2: bool
    q2_explaination: Optional[str]
    q3: bool
    q3_explaination: Optional[str]
    q4: bool
    q4_file_number: Optional[str]
    q4_explaination: Optional[str]
    q5: bool
    q5_explaination: Optional[str]
    q6: bool
    q6_explaination: Optional[str]
    q7: bool
    q7_explaination: Optional[str]


class Ee(BaseModel):
    ee_profile_no: str
    ee_expiry_date: date
    ee_jsvc: str
    ee_score: str
    ee_noc: str
    ee_job_title: str


class BcpnpModelApp(CommonModel):
    personal: Personal
    marriage: Marriage
    cor: List[COR]
    status: Status
    general: General
    joboffer: Joboffer
    family: List[Family]
    eraddress: List[ErAddress]
    contact: List[Contact]
    bcpnp: Bcpnp
    education: List[Education]
    employment: List[Employment]
    rcic: Rcic
    canadarelative: List[CanadaRelative]

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        if output_excel_file:
            excels = self.getExcels(
                ["excel/er.xlsx", "excel/pa.xlsx", "excel/bcpnp.xlsx", "excel/rep.xlsx"]
            )
        else:
            if excels is None or len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        super().__init__(excels, output_excel_file, globals())


class BcpnpEEModelApp(BcpnpModelApp):
    ee: Ee

    def __init__(self, excels=None, output_excel_file=None):
        super().__init__(excels=excels, output_excel_file=output_excel_file)



import xml.etree.ElementTree as ET
from source.tablesheet import TableNode
from source.excel import Excel
from model.common.utils import NameCodeConvert
from model.data.country import country
from model.data.language import language
from model.data.other import (
    phone_type,
    tr_canada_status,
    tr_marital_status,
    tr_form_study_level,
    tr_form_study_field,
    tr_portal_study_level,
    tr_portal_study_field,
    marital_stauts_5645,
)
from model.data.city import canada_city, canada_province


class PdfFormModel:
    """
    This model is for get information for TR vis or permit pdf xml, and save it to xlsx
    """

    def __init__(self, xml_file, model) -> None:
        self.data_model = model.get("data_model")
        self.convert_model = model.get("convert_model")
        self.remove_model = model.get("remove_model")
        self.tree = ET.ElementTree(file=xml_file)
        self.root = self.tree.getroot()
        self.data = (
            {}
        )  # data from pdf xml, and then foe excel file after some operations
        self._getData()  # get raw data from pdf xml
        self._convert()  # convert country, language, etc...
        self._assembl()  # use temp variables to assemble excel needed variables
        self._remove()  # remove not needed variables

    def _getSubItem(self, element, variable_xpath_pairs):
        temp_dict = {}
        for variable, xpath in variable_xpath_pairs.items():
            try:
                temp_dict[variable] = element.find(xpath).text
            except:
                print(f"{xpath} is invalid xpath")
        return temp_dict

    def _getData(self):
        for sheet, source_pair in self.data_model.items():
            if sheet not in ["occupation", "children", "siblings"]:
                self.data[sheet] = self._getSubItem(self.root, source_pair)
            else:
                self._getSpecial()

    def _getSpecial(self):
        pass

    # convert some items to excel format
    def _execConvert(self, pairs: dict, convert_item):
        for sheet, items in pairs.items():
            for item in items:
                # change value dict with converted country name
                sheet_data = self.data.get(sheet)
                if isinstance(sheet_data, list):
                    for i, sheet_item in enumerate(sheet_data):
                        code = sheet_data and sheet_data[i].get(item)
                        if code:
                            self.data[sheet][i][item] = NameCodeConvert(
                                convert_item
                            ).getName(code)
                else:
                    code = sheet_data and sheet_data.get(item)
                    if code:
                        self.data[sheet][item] = NameCodeConvert(convert_item).getName(
                            code
                        )

    def _convert(self):
        global_variables = globals()
        for pairs, model in self.convert_model.items():
            model = global_variables[
                pairs[:-6]
            ]  # get model by pairs name remove '_pairs'
            self._execConvert(self.convert_model[pairs], model)

        canada_city_pairs = {
            "sp": ["city"],
            "wp": ["employment_city"],
            "visa": ["city"],
            "vrincanada": ["city"],
            "spincanada": ["city"],
            "wpincanada": ["work_city"],
        }
        self._convert_city(canada_city_pairs, canada_city)

    def _convert_city(self, canada_city_pairs, canada_city):
        the_cities = {}
        for cities in canada_city.values():
            for city, code in cities.items():
                the_cities[city] = code
        self._execConvert(canada_city_pairs, the_cities)

    def _assembl(self):
        # for English or French
        self.data["personal"]["which_one_better"] = (
            "English" if self.data["personal"]["which_one_better"] == "01" else "French"
        )
        trcase_trcasein = "trcase" if "trcase" in self.data.keys() else "trcasein"
        self.data[trcase_trcasein]["service_in"] = (
            "English" if self.data[trcase_trcasein]["service_in"] == "01" else "French"
        )

        # assembly dob
        self.data["personal"]["dob"] = (
            self.data["personal"].get("dob_year")
            + "-"
            + self.data["personal"].get("dob_month")
            + "-"
            + self.data["personal"].get("dob_day")
        )
        dob_y, dob_m, dob_d = (
            self.data["marriage"].get("pre_sp_dob_year"),
            self.data["marriage"].get("pre_sp_dob_month"),
            self.data["marriage"].get("pre_sp_dob_day"),
        )
        self.data["marriage"]["pre_sp_dob"] = (
            dob_y + "-" + dob_m + "-" + dob_d if all([dob_y, dob_m, dob_d]) else None
        )

        self._cor_assemble()
        self._id_assemble()
        self._address_assemble()
        self._phone_assemble()
        # self._work_permit_assemble()
        self._education_assemble()
        self._occupation_assemble()

    # countries of residence
    def _cor_assemble(self):
        cor_dict = self.data["cor"]
        cor_current = {
            "country": cor_dict["current_cor_country"],
            "status": cor_dict["current_cor_status"],
            "other": cor_dict["current_cor_other"],
            "start_date": cor_dict["current_cor_start_date"],
            "end_date": cor_dict["current_cor_end_date"],
        }
        cor_pre1 = {
            "country": cor_dict["previous_cor_country1"],
            "status": cor_dict["previous_cor_status1"],
            "other": cor_dict["previous_cor_other1"],
            "start_date": cor_dict["previous_cor_start_date1"],
            "end_date": cor_dict["previous_cor_end_date1"],
        }
        cor_pre2 = {
            "country": cor_dict["previous_cor_country2"],
            "status": cor_dict["previous_cor_status2"],
            "other": cor_dict["previous_cor_other2"],
            "start_date": cor_dict["previous_cor_start_date2"],
            "end_date": cor_dict["previous_cor_end_date2"],
        }
        # re-sort the order of cor
        cor = []
        for c in [cor_current, cor_pre1, cor_pre2]:
            d = {}
            for key in ["start_date", "end_date", "country", "status"]:
                d[key] = c[key]
            cor.append(d)
        self.data["cor"] = cor

    # ids
    def _id_assemble(self):
        passport = {"variable_type": "passport", **self.data["passport"]}
        nid = {"variable_type": "id", **self.data["national_id"]}
        us_pr = {
            "variable_type": "pr",
            "number": self.data["us_pr"]["number"],
            "expiry_date": self.data["us_pr"]["expiry_date"],
        }
        self.data["personid"] = [passport, nid, us_pr]
        [self.data.pop(x) for x in ["passport", "national_id", "us_pr"]]

    # address
    def _address_assemble(self):
        m_dict = self.data["mailing_address"]
        r_dict = self.data["residential_address"]
        m = {"variable_type": "mailing_address", **m_dict}
        if r_dict["same_as_mailing"] == "Y":
            r = m
        else:
            r = {"variable_type": "residential_address", **r_dict}
        self.data["address"] = [m, r]
        [self.data.pop(x) for x in ["mailing_address", "residential_address"]]
        # convert country and province
        for i, addr in enumerate(self.data["address"]):
            self.data["address"][i]["country"] = NameCodeConvert(country).getName(
                addr["country"]
            )
            self.data["address"][i]["province"] = NameCodeConvert(
                canada_province
            ).getName(addr["province"])

    # phone
    def _phone_assemble(self):
        temp_phone = []
        for p_type in ["phone", "altphone", "fax"]:
            d = self.data[p_type]
            if d.get("number"):  # 如果没有号码，就位空
                d["variable_type"] = (
                    "fax" if not d.get("variable_type") else d["variable_type"]
                )
                temp_phone.append(
                    {
                        k: v
                        for k, v in d.items()
                        if k in ["variable_type", "country_code", "number", "ext"]
                    }
                )
        [self.data.pop(x) for x in ["phone", "altphone", "fax"]]
        self.data["phone"] = temp_phone

    # work permit TODO: 这个不应该在里面，应该是1295的特别的。
    def _work_permit_assemble(self):
        temp_wp = {}
        for d in [
            self.data["detailed_work"],
            self.data["intended_location"],
            self.data["intended_job"],
        ]:
            temp_wp = {**temp_wp, **{k: v for k, v in d.items()}}
        [
            self.data.pop(x)
            for x in ["detailed_work", "intended_location", "intended_job"]
        ]
        self.data["wp"] = temp_wp

    # education
    def _education_assemble(self):
        self.data["education"]["start_date"] = (
            self.data["education"]["start_date_year"]
            + "-"
            + self.data["education"]["start_date_month"]
            + "-01"
        )
        self.data["education"]["end_date"] = (
            self.data["education"]["end_date_year"]
            + "-"
            + self.data["education"]["end_date_month"]
            + "-01"
        )
        # re-sort education's order
        temp = {}
        for k in [
            "start_date",
            "end_date",
            "school_name",
            "education_level",
            "field_of_study",
            "city",
            "province",
            "country",
        ]:
            temp[k] = self.data.get("education").get(k, None)
        self.data["education"] = temp

    def _remove(self):
        for k, v in self.remove_model.items():
            [self.data[k].pop(item) for item in v]

    # occupation
    def _occupation_assemble(self):
        temp = []
        for occupation in self.data["occupations"]:
            occupation["start_date"] = (
                occupation["start_date_year"]
                + "-"
                + occupation["start_date_month"]
                + "-01"
                if all([occupation["start_date_year"], occupation["start_date_month"]])
                else ""
            )
            occupation["end_date"] = (
                occupation["end_date_year"] + "-" + occupation["end_date_month"] + "-01"
                if all([occupation["end_date_year"], occupation["end_date_month"]])
                else ""
            )
            occupation["country"] = NameCodeConvert(country).getName(
                occupation["country"]
            )  # convert country code to country name
            occupation["province"] = NameCodeConvert(canada_province).getName(
                occupation["province"]
            )  # convert province code to province name
            temp.append(occupation)
        temp1 = []
        for t in temp:
            d = {}
            for k in [
                "start_date",
                "end_date",
                "job_title",
                "company",
                "city",
                "province",
                "country",
            ]:
                d[k] = t[k]
            temp1.append(d)

        self.data["employment"] = temp1

    # get same xx_type
    def _getTableNodeWithType(self, table_node_list, the_type):
        try:
            for node in table_node_list:
                if node["variable_type"] == the_type:
                    return node
        except:
            return None

    # excel_file should be an existed excel file matching the 1294 program
    def _makeExcel(self, excel_file, protection):

        excel_obj = Excel(
            excel_file
        )  # if path.isfile(excel_file) else self.new_excel_obj

        for sheet_name, sheet_dict in self.data.items():
            # 如果不是sheet 下一个
            if sheet_name not in excel_obj.sheets.keys():
                continue

            for key, value in sheet_dict.items():
                # 如果excel表中有对应的sheet和字段
                if key in excel_obj.sheets[sheet_name].data.keys():
                    excel_obj.sheets[sheet_name].data[key].value = value

        for table_name, table_obj in excel_obj.tables.items():
            # this three tables will be processed independently
            if table_name in ["personid", "address", "phone"]:
                for index, value in enumerate(table_obj.data):
                    if value:
                        node = self._getTableNodeWithType(
                            self.data.get(table_name), value.variable_type
                        )
                        for k in value.__dict__.keys():
                            value = node.get(k) if node else None
                            if k not in ["variable_type", "display_type"]:
                                setattr(
                                    excel_obj.tables[table_name].data[index], k, value
                                )
            # others are normal
            else:
                # 现在在每张表里，取出sel.data[table_name]的table_node数据字典，做成tablenode,放到表里的data中。
                node_data = self.data.get(table_name)
                if isinstance(node_data, dict):
                    excel_obj.tables[table_name].data.append(TableNode(**node_data))
                elif isinstance(node_data, list):
                    for node in node_data:
                        excel_obj.tables[table_name].data.append(TableNode(**node))
                else:
                    pass
        # 根据最终更新了variables的sheets和talbes，生成新的excel文件
        excel_obj.makeExcel(excel_file, protection=protection)



from datetime import date
from typing import Optional, List
from pydantic import BaseModel, root_validator, EmailStr
from model.common.commonmodel import CommonModel
from model.common.address import Address
from model.common.contact import ContactBase
from termcolor import colored
from model.common.jobposition import PositionBase
from model.common.jobofferbase import JobofferBase
from model.common.person import Person


class General(BaseModel):
    cra_number: str
    recruit_email: EmailStr
    has_jobbank_account: bool
    has_bc_employer_certificate: bool


class ErAddress(Address):
    phone: str


class Contact(ContactBase):
    middle_name: Optional[str]
    position: str
    po_box: Optional[str]
    unit: Optional[str]
    street_number: str
    city: str
    province: str
    post_code: str

    @root_validator
    def checkCanadaProvince(cls, values):
        province = values["province"]
        if province not in [
            "AB",
            "BC",
            "MB",
            "NB",
            "NL",
            "NS",
            "NT",
            "NU",
            "ON",
            "PE",
            "QC",
            "SK",
            "YT",
        ]:
            raise ValueError(
                colored(
                    f'Since country is Canada is, the province must be one of  "AB","BC","MB","NB","NL","NS","NT","NU","ON","PE","QC","SK","YT"',
                    "red",
                )
            )
        return values


class Finance(BaseModel):
    year: int
    total_asset: float
    net_asset: float
    revenue: float
    net_income: float
    retained_earning: float


class Joboffer(JobofferBase):
    phone_country_code: str
    phone: str
    license_request: bool
    license_description: Optional[str]
    union: bool
    payment_way: str
    ot_after_hours_unit: Optional[str]
    ot_after_hours: Optional[float]
    is_working: bool
    has_probation: bool
    probation_duration: Optional[str]
    disability_insurance: bool
    dental_insurance: bool
    empolyer_provided_persion: bool
    extended_medical_insurance: bool
    extra_benefits: Optional[str]
    offer_date: date
    supervisor_name: str
    supervisor_title: str
    employer_rep: str
    employer_rep_title: str
    vacation_pay_weeks: int
    vacation_pay_percentage: float
    duties_brief: str
    duties: str
    english_french: bool
    oral: Optional[str]
    writing: Optional[str]
    reason_for_no: Optional[str]
    other_language_required: bool
    reason_for_other: Optional[str]
    education_level: str
    specific_edu_requirement: Optional[str]
    skill_experience_requirement: Optional[str]
    other_requirements: Optional[str]

    @root_validator
    def checkProbation(cls, values):
        has_probation = values.get("has_probation", None)
        probation_duration = values["probation_duration"]
        if has_probation and not probation_duration:
            raise ValueError(
                "Since has probation is true, but did not input probation duration"
            )
        return values

    @root_validator
    def checkLanguage(cls, values):
        english_french = values.get("english_french", None)
        oral = values.get("oral", None)
        writing = values.get("writing", None)
        reason_for_no = values.get("reason_for_no", None)
        if english_french and not oral or not writing:
            raise ValueError(
                "Since English or French is true, but did not input oral or/and writting language requirement"
            )
        if not english_french and not reason_for_no:
            raise ValueError(
                "Since there is no English or French requirement, but did not input the reason"
            )
        return values

    @root_validator
    def checkOtherLanguageReason(cls, values):
        other_language_required = values.get("other_language_required", None)
        reason_for_other = values.get("reason_for_other", None)
        if other_language_required and not reason_for_other:
            raise ValueError(
                "Since required other language in job offer sheet,but did not input the reason"
            )
        return values


class Lmi(BaseModel):
    brief_benefit: str
    job_creation_benefit: Optional[str]
    skill_transfer_benefit: Optional[str]
    fill_shortage_benefit: Optional[str]
    other_benefit: Optional[str]
    canadian_lost_job: bool
    canadian_lost_job_info: Optional[str]

    @root_validator
    def checkLabourDisput(cls, values):
        canadian_lost_job = values.get("canadian_lost_job", None)
        canadian_lost_job_info = values.get("canadian_lost_job_info", None)
        if canadian_lost_job and not canadian_lost_job_info:
            raise ValueError(
                "Since Canadian will lost job because of hiring TFW,but did not input the details about it."
            )
        return values


class Personal(Person):
    pass


class Position(PositionBase):
    pass


class LmiaRecruitment(CommonModel):
    general: General
    personal: Personal
    position: Position
    eraddress: List[ErAddress]
    contact: List[Contact]
    finance: List[Finance]
    joboffer: Joboffer
    lmi: Lmi
    person: Person
    address: List[Address]

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        if output_excel_file:
            excels = self.getExcels(
                ["excel/er.xlsx", "excel/lmia.xlsx", "excel/pa.xlsx"]
            )

        else:
            if excels is None and len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        # call parent class for validating
        super().__init__(excels, output_excel_file, globals())



from pydantic import BaseModel, validator
from datetime import date, timedelta
from typing import Optional, List
from model.common.mixins import DurationMixin, DatePeriod
from model.common.utils import makeList, Duration


class EmploymentBase(DatePeriod, DurationMixin):
    job_title: str
    noc_code: str
    weekly_hours: float
    company: str
    city: str
    province: Optional[str]
    country: str

    @property
    def part_time(self):
        return "Part time" if self.weekly_hours < 30 else ""

    @property
    def is_current(self):
        return True if self.end_date == "Present" else False

    @property
    def years(self):
        return Duration(self.start_date, self.end_date).years

    @property
    def months(self):
        return Duration(self.start_date, self.end_date).years

    def yearsOnDate(self, end_date):
        return Duration(self.start_date, self.end_date).yearsOnDate(end_date)

    def monthsOnDate(self, end_date):
        return Duration(self.start_date, self.end_date).monthsOnDate(end_date)

    @property
    def start_date_mmyyyy(self):
        return self.start_date.strftime("%b %Y")

    @property
    def end_date_mmyyyy(self):
        return self.end_date.strftime("%b %Y") if self.end_date else "Present"

    @property
    def is_present(self):
        return not self.end_date or self.end_date == "Present"



from pydantic import BaseModel, EmailStr, root_validator
from typing import Optional, List
from model.common.utils import checkRow


class ContactBase(BaseModel):
    variable_type: Optional[str]
    display_type: Optional[str]
    last_name: Optional[str]
    first_name: Optional[str]
    phone: Optional[str]
    email: Optional[EmailStr]

    class Config:
        anystr_lower = True

    @root_validator
    def checkCompletion(cls, values):
        all_fields = ["last_name", "first_name", "phone", "email"]
        required_fields = ["last_name", "first_name", "phone", "email"]
        checkRow(values, all_fields, required_fields)
        return values


class Contacts(object):
    def __init__(self, contacts: List[ContactBase]):
        self.contacts = contacts

    def _specific_contact(self, v_type):
        contact = [
            contact for contact in self.contacts if contact.variable_type == v_type
        ]
        return contact[0] if contact else None

    @property
    def primary(self):
        return self._specific_contact("primary")

    @property
    def second(self):
        return self._specific_contact("second")

    @property
    def preferredContact(self):
        for contact_type in ["primary", "second"]:
            contact = self._specific_contact(contact_type)
            if contact:
                return contact



from pydantic import BaseModel
from termcolor import colored
from functools import reduce
from typing import List, Optional
from model.common.address import Address
from model.common.phone import Phone
from model.common.trperson import (
    COR,
    PersonId,
    Personal,
    Marriage,
    Education,
    Employment,
)
from model.common.tr import TrCaseIn, TrBackground, VrInCanada
from model.common.commonmodel import CommonModel
from datetime import date


class M5708Model(CommonModel):
    personal: Personal
    marriage: Marriage
    personid: List[PersonId]
    address: List[Address]
    education: List[Education]
    employment: List[Employment]
    phone: List[Phone]
    cor: List[COR]
    trcasein: TrCaseIn
    vrincanada: VrInCanada
    trbackground: TrBackground

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        if output_excel_file:
            excels = self.getExcels(["excel/tr.xlsx", "excel/pa.xlsx"])
        else:
            if excels is None and len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        super().__init__(excels, output_excel_file, globals())

    # TODO: python-docx-template cannot use function returned value
    @property
    def visit_description(self):
        last_visit = self.trcasein.most_recent_entry_date
        original_visit = self.trcasein.original_entry_date
        last_place = self.trcasein.most_recent_entry_place
        original_place = self.trcasein.original_entry_place

        # if last visit existed, calculate the days
        visit_date = last_visit or original_visit
        visit_place = last_place or original_place
        days = (date.today() - visit_date).days

        apply_for = "restoration" if days > 180 else "extension"
        return f"I am applying for {apply_for}"

    @property
    def getAllDict(self):
        return {
            **self.dict(),
            "visit_description": self.visit_description,
            "subject": self.personal.subject,
            "object": self.personal.object,
            "attributive": self.personal.attributive,
            "salutation": self.personal.salutation,
            "short_name": self.personal.short_name,
        }



"""
This is a tool set for pydantic validation shared by all models
"""
from datetime import date
from dateutil.parser import parse, ParserError
from typing import List


def normalize(name: str):
    if name:
        return " ".join((word.capitalize()) for word in name.split(" "))
    return ""


def makeList(name: str):
    if name:
        return name.split("\n")
    return []


def trimString(text: str):
    return "" if text == None else text.strip()


class Duration:
    def __init__(self, start_date, end_date):
        try:
            self.start_date = (
                start_date if isinstance(start_date, date) else parse(start_date)
            )
            self.end_date = end_date if isinstance(end_date, date) else parse(end_date)
        except ParserError as err:
            raise ParserError(f'{err.args[0]}," in file ",{__file__}')

    @property
    def years(self):
        return (
            self.end_date.year
            - self.start_date.year
            - (
                (self.end_date.month, self.end_date.day)
                < (self.start_date.month, self.start_date.day)
            )
        )

    @property
    def months(self):
        years2months = (self.end_date.year - self.start_date.year) * 12
        months2months = self.end_date.month - self.start_date.month
        return years2months + months2months

    # get years on a specific date
    def yearsOnDate(self, end_date=date.today()):
        if not isinstance(end_date, date):
            try:
                end_date = parse(end_date)
                return (
                    end_date.year
                    - self.start_date.year
                    - (
                        (end_date.month, end_date.day)
                        < (self.start_date.month, self.start_date.day)
                    )
                )
            except ParserError as err:
                raise ParserError(f'{err.args[0]}," in file ",{__file__}')
        return (
            end_date.year
            - self.start_date.year
            - (
                (end_date.month, end_date.day)
                < (self.start_date.month, self.start_date.day)
            )
        )

    # get months on a specific date
    def monthsOnDate(self, end_date=date.today()):
        try:
            end_date = end_date if isinstance(end_date, date) else parse(end_date)
        except ParserError as err:
            raise ParserError(f'{err.args[0]}," in file ",{__file__}')

        years2months = (end_date.year - self.start_date.year) * 12
        months2months = end_date.month - self.start_date.month
        return years2months + months2months

    # get days
    @property
    def days(self):
        return (self.end_date - self.start_date).days

    def daysOnDate(self, end_date=date.today()):
        if not isinstance(end_date, date):
            try:
                end_date = parse(end_date).date()
                return (end_date - self.start_date).days
            except ParserError as err:
                raise ParserError(f'{err.args[0]}," in file ",{__file__}')
        return (end_date - self.start_date).days


# Used in pydantic class to check if a row has input but missed some fields, values is root_validator values
def checkRow(values: dict, all_fields: List[str], required_fields: List[str]):
    all_fields_values = [values.get(field, None) for field in all_fields]
    required_values = [values.get(field, None) for field in required_fields]
    has_values = [value for value in required_values if value]

    if any(all_fields_values) and not all(required_values):
        raise ValueError(
            f"Please check the row with values ({has_values}), some required fileds are missed."
        )


"""
This class is for code and name convert. Usually used for country, province,city, etc... to convert between name and code
"""


class NameCodeConvert(object):
    # initialize with a dict including name and code
    def __init__(self, d: dict):
        self.d = d

    def getName(self, code):
        c = {v: k for k, v in self.d.items()}
        return c.get(str(code))

    def getCode(self, name):
        # remove additional space and trim left and right
        name = " ".join(name.split()).strip()
        for item_name, code in self.d.items():
            if name in item_name.lower():
                return code


# TODO: 此处可以考虑增加模糊选项，如果有错，提供对应的输入建议。 参考noc的输入检查


def main():
    d = Duration("2020-01-01", "2022-01-02")
    print(d.days)


if __name__ == "__main__":
    main()



from pydantic import BaseModel
from model.common.commonmodel import CommonModel
import os

from pydantic import BaseModel
from model.common.advertisement import Advertisement,InterviewRecord

class RecruitmentSummary(BaseModel):
    apply_email:bool
    interview_process_evidence:bool
    joboffer_email:bool
    joboffer_email_reply:bool
    
class LmiaApplication(CommonModel):
    
    advertisement:Advertisement
    interviewrecord:InterviewRecord
    recruitmentsummary:RecruitmentSummary
        
    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self,excels=None,output_excel_file=None):
        if output_excel_file:
            path=os.path.abspath(os.path.join(os.path.dirname(__file__), os.path.pardir))
            excels=self.getExcels(['excel/recruitment.xlsx'])
        else:
            if excels is None and len(excels)==0:
                raise ValueError('You must input excel file list as source data for validation')
        # call parent class for validating
        super().__init__(excels,output_excel_file,globals())




# before '_' is sheet name
# variable as key, xpath of xml as value.
data_model = {
    "trcase": {
        "service_in": "Page1/PersonalDetails/ServiceIn/ServiceIn",
        "same_as_cor": "Page1/PersonalDetails/SameAsCORIndicator",
        "applying_country": "Page1/PersonalDetails/CountryWhereApplying/Row2/Country",
        "applying_stauts": "Page1/PersonalDetails/CountryWhereApplying/Row2/Status",
        "applying_start_date": "Page1/PersonalDetails/CountryWhereApplying/Row2/FromDate",
        "applying_end_date": "Page1/PersonalDetails/CountryWhereApplying/Row2/ToDate",
        "consent_of_info_release": "Page4/Consent0/Choice",
        "submission_date": "Page4/Consent0/C1CertificateIssueDate",
    },
    "personal": {
        "uci": "Page1/PersonalDetails/UCIClientID",
        "last_name": "Page1/PersonalDetails/Name/FamilyName",
        "first_name": "Page1/PersonalDetails/Name/GivenName",
        "has_alias_name": "Page1/PersonalDetails/AliasName/AliasNameIndicator/AliasNameIndicator",
        "used_last_name": "Page1/PersonalDetails/AliasName/AliasFamilyName",
        "used_first_name": "Page1/PersonalDetails/AliasName/AliasGivenName",
        "sex": "Page1/PersonalDetails/Sex/Sex",
        "dob_year": "Page1/PersonalDetails/DOBYear",
        "dob_month": "Page1/PersonalDetails/DOBMonth",
        "dob_day": "Page1/PersonalDetails/DOBDay",
        "place_of_birth": "Page1/PersonalDetails/PlaceBirthCity",
        "country_of_birth": "Page1/PersonalDetails/PlaceBirthCountry",
        "citizen": "Page1/PersonalDetails/Citizenship/Citizenship",
        "native_language": "Page2/MaritalStatus/SectionA/Languages/languages/nativeLang/nativeLang",
        "english_french": "Page2/MaritalStatus/SectionA/Languages/languages/ableToCommunicate/ableToCommunicate",
        "which_one_better": "Page2/MaritalStatus/SectionA/Languages/languages/lov",
        "language_test": "Page2/MaritalStatus/SectionA/Languages/LanguageTest",
        "email": "Page3/FaxEmail/Email",
    },
    # country of residence: current and past for over 6 months. table model
    "cor": {
        "current_cor_country": "Page1/PersonalDetails/CurrentCOR/Row2/Country",
        "current_cor_status": "Page1/PersonalDetails/CurrentCOR/Row2/Status",
        "current_cor_other": "Page1/PersonalDetails/CurrentCOR/Row2/Other",
        "current_cor_start_date": "Page1/PersonalDetails/CurrentCOR/Row2/FromDate",
        "current_cor_end_date": "Page1/PersonalDetails/CurrentCOR/Row2/ToDate",
        "previous_cor_country1": "Page1/PersonalDetails/PreviousCOR/Row2/Country",
        "previous_cor_status1": "Page1/PersonalDetails/PreviousCOR/Row2/Status",
        "previous_cor_other1": "Page1/PersonalDetails/PreviousCOR/Row2/Other",
        "previous_cor_start_date1": "Page1/PersonalDetails/PreviousCOR/Row2/FromDate",
        "previous_cor_end_date1": "Page1/PersonalDetails/PreviousCOR/Row2/ToDate",
        "previous_cor_country2": "Page1/PersonalDetails/PreviousCOR/Row3/Country",
        "previous_cor_status2": "Page1/PersonalDetails/PreviousCOR/Row3/Status",
        "previous_cor_other2": "Page1/PersonalDetails/PreviousCOR/Row3/Other",
        "previous_cor_start_date2": "Page1/PersonalDetails/PreviousCOR/Row3/FromDate",
        "previous_cor_end_date2": "Page1/PersonalDetails/PreviousCOR/Row3/ToDate",
    },
    # marriage
    "marriage": {
        "marital_status": "Page1/MaritalStatus/SectionA/MaritalStatus",
        "married_date": "Page1/MaritalStatus/SectionA/DateOfMarriage",
        "sp_last_name": "Page1/MaritalStatus/SectionA/FamilyName",
        "sp_first_name": "Page1/MaritalStatus/SectionA/GivenName",
        "previous_married": "Page2/MaritalStatus/SectionA/PrevMarriedIndicator",
        "pre_sp_last_name": "Page2/MaritalStatus/SectionA/PMFamilyName",
        "pre_sp_first_name": "Page2/MaritalStatus/SectionA/GivenName/PMGivenName",
        "pre_relationship_type": "Page2/MaritalStatus/SectionA/TypeOfRelationship",
        "pre_sp_dob_year": "Page2/MaritalStatus/SectionA/PrevSpouseDOB/DOBYear",
        "pre_sp_dob_month": "Page2/MaritalStatus/SectionA/PrevSpouseDOB/DOBMonth",
        "pre_sp_dob_day": "Page2/MaritalStatus/SectionA/PrevSpouseDOB/DOBDay",
        "pre_start_date": "Page2/MaritalStatus/SectionA/FromDate",
        "pre_end_date": "Page2/MaritalStatus/SectionA/ToDate/ToDate",
    },
    "passport": {
        "number": "Page2/MaritalStatus/SectionA/Passport/PassportNum/PassportNum",
        "country": "Page2/MaritalStatus/SectionA/Passport/CountryofIssue/CountryofIssue",
        "issue_date": "Page2/MaritalStatus/SectionA/Passport/IssueDate/IssueDate",
        "expiry_date": "Page2/MaritalStatus/SectionA/Passport/ExpiryDate",
    },
    "national_id": {
        "has_id": "Page2/natID/q1/natIDIndicator",
        "number": "Page2/natID/natIDdocs/DocNum/DocNum",
        "country": "Page2/natID/natIDdocs/CountryofIssue/CountryofIssue",
        "issue_date": "Page2/natID/natIDdocs/IssueDate/IssueDate",
        "expiry_date": "Page2/natID/natIDdocs/ExpiryDate",
    },
    "us_pr": {
        "has_id": "Page2/USCard/q1/usCardIndicator",
        "number": "Page2/USCard/usCarddocs/DocNum/DocNum",
        "expiry_date": "Page2/USCard/usCarddocs/ExpiryDate",
    },
    "mailing_address": {
        "po_box": "Page2/ContactInformation/contact/AddressRow1/POBox/POBox",
        "unit": "Page2/ContactInformation/contact/AddressRow1/Apt/AptUnit",
        "street_number": "Page2/ContactInformation/contact/AddressRow1/StreetNum/StreetNum",
        "street_name": "Page2/ContactInformation/contact/AddressRow1/Streetname/Streetname",
        "city": "Page2/ContactInformation/contact/AddressRow2/CityTow/CityTown",
        "country": "Page2/ContactInformation/contact/AddressRow2/Country/Country",
        "province": "Page2/ContactInformation/contact/AddressRow2/ProvinceState/ProvinceState",
        "post_code": "Page2/ContactInformation/contact/AddressRow2/PostalCode/PostalCode",
        "district": "Page2/ContactInformation/contact/AddressRow2/District",
    },
    "residential_address": {
        "same_as_mailing": "Page2/ContactInformation/contact/SameAsMailingIndicator",
        "unit": "Page2/ContactInformation/contact/ResidentialAddressRow1/AptUnit/AptUnit",
        "street_number": "Page2/ContactInformation/contact/ResidentialAddressRow1/StreetNum/StreetNum",
        "street_name": "Page2/ContactInformation/contact/ResidentialAddressRow1/StreetName/Streetname",
        "city": "Page2/ContactInformation/contact/ResidentialAddressRow1/CityTown/CityTown",
        "country": "Page2/ContactInformation/contact/ResidentialAddressRow2/Country/Country",
        "province": "Page2/ContactInformation/contact/ResidentialAddressRow2/ProvinceState/ProvinceState",
        "post_code": "Page2/ContactInformation/contact/ResidentialAddressRow2/PostalCode/PostalCode",
        "district": "Page2/ContactInformation/contact/ResidentialAddressRow2/District",
    },
    "phone": {
        "variable_type": "Page2/ContactInformation/contact/PhoneNumbers/Phone/Type",
        "canada_us": "Page2/ContactInformation/contact/PhoneNumbers/Phone/CanadaUS",
        "other": "Page2/ContactInformation/contact/PhoneNumbers/Phone/Other",
        "ext": "Page2/ContactInformation/contact/PhoneNumbers/Phone/NumberExt",
        "country_code": "Page2/ContactInformation/contact/PhoneNumbers/Phone/NumberCountry",
        "number": "Page2/ContactInformation/contact/PhoneNumbers/Phone/ActualNumber",
        "na_area": "Page2/ContactInformation/contact/PhoneNumbers/Phone/NANumber/AreaCode",
        "na_first_three": "Page2/ContactInformation/contact/PhoneNumbers/Phone/NANumber/FirstThree",
        "na_last_five": "Page2/ContactInformation/contact/PhoneNumbers/Phone/NANumber/LastFive",
        "int_number": "Page2/ContactInformation/contact/PhoneNumbers/Phone/IntlNumber/IntlNumber",
    },
    "altphone": {
        "variable_type": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/Type",
        "canada_us": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/CanadaUS",
        "other": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/Other",
        "ext": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NumberExt",
        "country_code": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NumberCountry",
        "number": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/ActualNumber",
        "na_area": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NANumber/AreaCode",
        "na_first_three": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NANumber/FirstThree",
        "na_last_five": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/NANumber/LastFive",
        "int_number": "Page2/ContactInformation/contact/PhoneNumbers/AltPhone/IntlNumber/IntlNumber",
    },
    "fax": {
        "canada_us": "Page3/FaxEmail/Phone/CanadaUS",
        "other": "Page3/FaxEmail/Phone/Other",
        "ext": "Page3/FaxEmail/Phone/NumberExt",
        "country_code": "Page3/FaxEmail/Phone/NumberCountry",
        "number": "Page3/FaxEmail/Phone/ActualNumber",
        "na_area": "Page3/FaxEmail/Phone/NANumber/AreaCode",
        "na_first_three": "Page3/FaxEmail/Phone/NANumber/FirstThree",
        "na_last_five": "Page3/FaxEmail/Phone/NANumber/LastFive",
        "int_number": "Page3/FaxEmail/Phone/IntlNumber/IntlNumber",
    },
    "wp": {
        "work_permit_type": "Page3/DetailsOfIntendedWork/DetailsOfWork/TypeofWork/WorkPermitType",
        "employer_name": "Page3/DetailsOfIntendedWork/DetailsOfWork/PurposeRow1/EmployerName/EmployerName",
        "employer_address": "Page3/DetailsOfIntendedWork/DetailsOfWork/PurposeRow1/Address/Address",
        "employment_province": "Page3/IntendedLocationInCanada/intendedLocation/ProvinceState/ProvinceState",
        "employment_city": "Page3/IntendedLocationInCanada/intendedLocation/CityTown/CityTown",
        "work_location_address": "Page3/IntendedLocationInCanada/intendedLocation/Address",
        "job_title": "Page3/DetailsOfWorkCont/details/jobTitle",
        "brief_duties": "Page3/DetailsOfWorkCont/details/posDesc",
        "start_date": "Page3/DetailsOfWorkCont/details/HowLongStudy/FromDate",
        "end_date": "Page3/DetailsOfWorkCont/details/HowLongStudy/ToDate",
        "lmia_num_or_offer_num": "Page3/DetailsOfWorkCont/details/LMO/LMO",
    },
    "education": {
        "start_date_year": "Page3/PageWrapper/Education/Edu_Row1/FromYear",
        "start_date_month": "Page3/PageWrapper/Education/Edu_Row1/FromMonth",
        "end_date_year": "Page3/PageWrapper/Education/Edu_Row1/ToYear",
        "end_date_month": "Page3/PageWrapper/Education/Edu_Row1/ToMonth",
        "school_name": "Page3/PageWrapper/Education/Edu_Row1/School",
        "field_of_study": "Page3/PageWrapper/Education/Edu_Row1/FieldOfStudy",
        "city": "Page3/PageWrapper/Education/Edu_Row1/CityTown",
        "country": "Page3/PageWrapper/Education/Edu_Row1/Country/Country",
        "province": "Page3/PageWrapper/Education/Edu_Row1/ProvState",
    },
    # occupation is a special dict, and will be handled seperately
    "occupation": {
        "start_date_year": "FromYear",
        "start_date_month": "FromMonth",
        "end_date_year": "ToYear",
        "end_date_month": "ToMonth",
        "job_title": "Occupation/Occupation",
        "company": "Employer",
        "city": "CityTown/CityTown",
        "province": "ProvState",
        "country": "Country/Country",
    },
    "trbackground": {
        "q1a": "Page4/BackgroundInfo/Choice",
        "q1b": "Page4/BackgroundInfo/Choice",
        "q1c": "Page4/BackgroundInfo/Details/MedicalDetails",
        "q2a": "Page4/BackgroundInfo2/VisaChoice1",
        "q2b": "Page4/BackgroundInfo2/VisaChoice2",
        "q2c": "Page4/BackgroundInfo2/VisaChoice3",
        "q2d": "Page4/BackgroundInfo2/Details/refusedDetails",
        "q3a": "Page4/PageWrapper/BackgroundInfo3/Choice",
        "q3b": "Page4/PageWrapper/BackgroundInfo3/militaryServiceDetails",
        "q4a": "Page4/PageWrapper/Military/Choice",
        "q4b": "Page4/PageWrapper/Military/militaryServiceDetails",
        "q5": "Page4/PageWrapper/Occupation/Choice",
        "q6": "Page4/PageWrapper/GovPosition/Choice",
    },
}

convert_model = {
    "country_pairs": {
        "trcase": ["applying_country"],
        "personal": ["country_of_birth", "citizen"],
        "passport": ["country"],
        "national_id": ["country"],
        "education": ["country"],
        "cor": [
            "current_cor_country",
            "previous_cor_country1",
            "previous_cor_country2",
        ],
    },
    "canada_province_pairs": {
        "wp": ["employment_province"],
        "address": ["province"],
        "education": ["province"],
        "employment": ["province"],
    },
    "language_pairs": {"personal": ["native_language"]},
    "phone_type_pairs": {"phone": ["variable_type"], "altphone": ["variable_type"]},
    "tr_canada_status_pairs": {
        "cor": ["current_cor_status", "previous_cor_status1", "previous_cor_status2"]
    },
    "tr_marital_status_pairs": {
        "marriage": ["marital_status", "pre_relationship_type"]
    },
}

remove_model = {
    "personal": ["dob_year", "dob_month", "dob_day"],
    "marriage": ["pre_sp_dob_year", "pre_sp_dob_month", "pre_sp_dob_day"],
}



from datetime import date
from pydantic import BaseModel, EmailStr, validator
from typing import Optional
from model.common.educationbase import EducationBase
from model.common.person import Person
from model.common.utils import normalize, checkRow
from pydantic.class_validators import root_validator
from model.common.mixins import DatePeriod
from model.data.city import canada_province


class Personal(Person):
    used_last_name: Optional[str]
    used_first_name: Optional[str]
    uci: Optional[str]
    country_of_birth: str
    place_of_birth: str
    email: EmailStr

    native_language: str
    english_french: str
    which_one_better: Optional[str]
    language_test: bool

    _normalize_used_first_name = validator(
        "used_first_name", allow_reuse=True, check_fields=False
    )(normalize)
    _normalize_used_last_name = validator(
        "used_last_name", allow_reuse=True, check_fields=False
    )(normalize)

    @root_validator
    def checkAnswers(cls, values):
        questions = ["english_french"]
        explanations = ["which_one_better"]
        qas = dict(zip(questions, explanations))
        for k, v in qas.items():
            if values.get(k) == "Both" and not values.get(v):
                raise ValueError(
                    f"Since {k} is true, but you did not answer the question {v} in info-position sheet"
                )
        return values


class Marriage(BaseModel):
    marital_status: str
    married_date: Optional[date]
    sp_last_name: Optional[str]
    sp_first_name: Optional[str]
    sp_is_canadian: Optional[bool]
    previous_married: bool
    pre_sp_last_name: Optional[str]
    pre_sp_first_name: Optional[str]
    pre_relationship_type: Optional[str]
    pre_sp_dob: Optional[date]
    pre_start_date: Optional[date]
    pre_end_date: Optional[date]


class PersonId(BaseModel):
    variable_type: str
    display_type: str
    number: Optional[str]
    country: Optional[str]
    issue_date: Optional[date]
    expiry_date: Optional[date]

    @root_validator
    def checkCompletion(cls, values):
        all_fields = ["number", "country", "issue_date", "expiry_date"]
        required_fields = ["number", "country", "issue_date", "expiry_date"]
        checkRow(values, all_fields, required_fields)
        return values


# Not everyone has education. So, it's optional. Without start date and end date, the app will regard having no education
class Education(EducationBase):
    city: Optional[str]
    province: Optional[str]
    country: Optional[str]

    @root_validator
    def checkCompletion(cls, values):
        all_fields = [
            "start_date",
            "end_date",
            "school_name",
            "education_level",
            "field_of_study",
            "city",
            "province",
            "country",
        ]
        required_fields = [
            "start_date",
            "school_name",
            "education_level",
            "field_of_study",
            "city",
            "country",
        ]
        checkRow(values, all_fields, required_fields)
        return values

    @root_validator
    def checkProvince(cls, values):
        country = values.get("country")
        province = values.get("province")
        if country and country.lower() == "canada":
            if not province:
                raise ValueError("When the country is Canada, you must input province")
            if province not in canada_province.keys():
                raise ValueError(
                    f"Canada province must be one of {canada_province.keys()}"
                )
        return values


# Not everyone has employment experience. So, it's optional. Without start date and end date, the app will regard having no post secondary education
class Employment(DatePeriod):
    job_title: Optional[str]
    company: Optional[str]
    city: Optional[str]
    province: Optional[str]
    country: Optional[str]

    @root_validator
    def checkCompletion(cls, values):
        all_fields = [
            "start_date",
            "end_date",
            "job_title",
            "company",
            "city",
            "province",
            "country",
        ]
        required_fields = [
            "start_date",
            "job_title",
            "company",
            "city",
            "country",
        ]
        checkRow(values, all_fields, required_fields)
        return values


class Travel(DatePeriod):
    length: Optional[int]
    destination: Optional[str]
    purpose: Optional[str]

    @root_validator
    def checkCompletion(cls, values):
        all_fields = ["start_date", "end_date", "length", "destination", "purpose"]
        required_fields = ["start_date", "end_date", "length", "destination", "purpose"]
        checkRow(values, all_fields, required_fields)
        return values


class Family(BaseModel):
    last_name: str
    first_name: str
    native_last_name: str
    native_first_name: str
    marital_status: str
    date_of_birth: date
    birth_country: str
    address: str
    occupation: Optional[str]
    relationship: str
    email: Optional[EmailStr]
    date_of_death: Optional[date]
    place_of_death: Optional[str]
    accompany_to_canada: bool


# Countries of Residence
class COR(DatePeriod):
    country: str
    status: str



from functools import reduce
from pydantic import BaseModel, validator
from datetime import date
from model.common.utils import Duration
from typing import Optional

# Compare if self obj and another obj is same with compare all property and its value
class EqMixin(object):
    def __eq__(self, another):
        for k, v in self.__dict__.items():
            if v != getattr(another, k, None):
                return False
        return True


class KwargsMixin(object):
    def _getInput(self, global_dict, excel_obj):
        """
        Prerequisite of this mixin
        1. excel_obj is the one that all input excels combined, no sheet has same name
        2. Final level model is constructed with all sheets/tables, no specific single varialbe
        3. Only InfoSheet and TableList two data structure in the final level model
        """
        input_kwargs = {}
        # According to the prerequisite described above, all the referrenced scheme are defined in 'definitions'.
        # key is Model name, and after lower(), it is also excel sheet name and the property defined in this class. They are same. This is the requirement.
        for k in self.schema()["definitions"].keys():
            # globals() returns a dict inlcuding all global classes, methods, ....
            klass = global_dict.get(k)
            # use k.lower (it's a sheet name here) to get sheet data
            sheet_data = excel_obj.dict.get(k.lower(), None)
            # only two types, one is list of model objects,another is model object
            if not isinstance(sheet_data, list):
                # assign the input kwargs. Here k.lower() is class's property. klass is the model, on which the app create an obj
                input_kwargs[k.lower()] = klass(**sheet_data)
            else:
                input_kwargs[k.lower()] = [klass(**d) for d in sheet_data]
        return input_kwargs


class MakeExcelMixin(object):
    def _generateExcel(self, global_dict, excel_obj, excel_file):
        sheets = []
        tables = []
        # 获得每一sheet对应class，并获得它的properties list
        for k in self.schema()["definitions"].keys():
            # globals() returns a dict inlcuding all global classes, methods, ....
            klass = global_dict.get(k)  # get the class
            variables = list(
                klass.__fields__.keys()
            )  # get all variables in the shhet or table
            sheet_table_name = (
                k.lower()
            )  # k.lower is the sheet / table name according to convention
            if "info-" + sheet_table_name in excel_obj.sheet_names:
                sheets.append(
                    excel_obj.getSheet(sheet_table_name, variables)
                )  # get sheet object
            if "table-" + sheet_table_name in excel_obj.table_names:
                tables.append(
                    excel_obj.getTable(sheet_table_name, variables)
                )  # get table object

        # 根据最终更新了variables的sheets和talbes，生成新的excel文件
        excel_obj.makeExcel(excel_file, sheets=sheets, tables=tables)


# 处理所包含start_date 和 end_date的记录。
class DurationMixin(object):
    @property
    def start_from(self):
        return self.start_date.strftime("%b %Y")

    @property
    def end_to(self):
        return self.end_date.strftime("%b %Y") if self.end_date else "Present"

    @property
    def lengthOfYears(self):
        return Duration(self.start_date, self.end_date).years


# Check if end date is earlier than start date
class DatePeriod(BaseModel):
    start_date: date
    end_date: Optional[date]

    @validator("end_date")
    def endDateBigger(cls, end_date, values):
        start_date = values.get("start_date")
        if not start_date:
            return end_date
        the_date = end_date or date.today()
        if (the_date - start_date).days <= 0:
            raise ValueError(
                f"End date {the_date} is earlier than start date {start_date}"
                if end_date
                else f"{start_date} is later than today"
            )
        return end_date



from webform.definition import Action
from datetime import date

# specially processing date input, with or without enter after inputing date
def inputDate(
    label: str, id: str, value: date, required=True, length=10, with_enter=False
):
    the_date = [
        {
            "action_type": Action.Input.value,
            "label": label,
            "id": id,
            "value": value.strftime("%Y-%m-%d") if value else None,
            "required": required,
            "length": length,
        }
    ]
    the_enter = (
        [{"action_type": Action.PressKey.value, "label": label, "key": "Enter"}]
        if with_enter
        else []
    )

    return the_date + the_enter


def pressEnter():
    return {
        "action_type": Action.PressKey.value,
        "label": "Enter",
        "key": "Enter",
    }



#!/Users/jacky/tools/venv/bin/python3.9
import json
import copy
from functools import total_ordering
from collections import OrderedDict
from termcolor import colored


class TableNode(object):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        for k, v in kwargs.items():
            setattr(self, k, v)

    def __len__(self):
        return len(self.__dict__)

    # return True if every key and value is same and with same length
    def __eq__(self, another: object) -> bool:
        if len(self) != len(another):
            return False
        for p in self.__dict__:  # check every property's value is same
            if self.__dict__.get(p) != another.__dict__.get(p):
                return False
        return True

    # Return True if self and another has top three values in same, which we regard they are some record
    # The purpose of doing so is to check if two records are actually same, so has to be merged.
    def is_same_record(self, another: object):
        obj = copy.deepcopy(self)
        try:
            top_three = list(obj.__dict__)[0:3]
        except:
            raise IndexError(f"{self} has less than 3 columns")
        for property in top_three:
            if getattr(obj, property) != getattr(another, property):
                return False
        return True

    # + operator： if another has value, no matter self has value or not,  replace self’s value with another’s value, but if another has no value, keep self’s value.
    def __add__(self, another: object):
        obj = copy.deepcopy(self)
        # 相同变量部分：搜索another每一个值，如果不为空，就替代self。
        # 不同变量部分，another多出来的部分,增加变量和值到self。
        for p in another.__dict__:
            if p in self.__dict__:
                # If another has value, no matter self has value or not,  replace self’s value with another’s value
                if getattr(another, p, None):
                    setattr(obj, p, getattr(another, p))
                # but if another has no value, keep self’s value.
            else:
                # if a property in another but not in self, add it to self
                setattr(obj, p, getattr(another, p, None))

        return obj

    # copy:  if another has value, no matter self has value or not,  replace self’s value with another’s value
    def copy(self, another: object):
        obj = copy.deepcopy(self)
        # 相同变量部分：搜索another每一个值，替代self。
        for p in another.__dict__:
            if p in self.__dict__:
                setattr(obj, p, getattr(another, p))
        return obj

    def __hash__(self) -> int:
        hs = []
        for key, value in self.__dict__.items():
            hs.append("{} = {}".format(key, value))
        return hash(",".join(hs))

    # return the variable name representing the object
    def __str__(self):
        str_str = ["{}:{}".format(key, value) for key, value in self.__dict__.items()]
        return ", ".join(str_str)

    # return the string representing the class constructor method
    def __repr__(self):
        repr_str = [
            '{} = "{}"'.format(key, value)
            if type(value) == str
            else "{} = {}".format(key, value)
            for key, value in self.__dict__.items()
        ]
        return f'TableNode({",".join(repr_str)})'

    # return True or False representing node is null or not null.
    # null means the infoNode has no any value, reflecting in excel is a blank row
    def __bool__(self):
        return any(self.__dict__.values())


# sheet data as a dict xx
@total_ordering
class TableList(object):
    """[TableList class used for handle inside  of table sheet and between of table sheets level information]

    Args:
        2D list including three  rows of sheet title list, column title list, and variables list
        columns are defined by excel files.

    Raises:
        StopIteration: [description]
        KeyError: ['description']

    Returns:
        TableList object
    """

    def __init__(self, rows) -> None:
        super().__init__()
        self.rows = rows
        self.sheet_title = self._getSheetTitle()  # the first line is the sheet's title
        self.column_titles = self.rows.pop(0)  # the second line is the display title
        # self.column_variables=[self._title2variable(x) for x in variables]# the third line in rows is this sheet's schema, used as title variables
        self.column_variables = self._cleanTitles()
        self.column_titles = OrderedDict(
            zip(self.column_variables, self.column_titles)
        )  # 绑定title到variable，否则在后续+ - table的时候，title和variable 以及value可能无法保持一致。
        self.rows = rows  # data rows
        self.data = []  # contains the sheet's all TableNode objects
        self.createTableNodes(rows)  # rows after poped three rows
        self.index = 0

    def _getSheetTitle(self):
        st = self.rows.pop(0)
        st = st[0].split("\n")[0] if len(st) > 0 else "A TableList object"
        return st

    def _cleanTitles(self):
        raw_variables = self.rows.pop(0)
        variables = [x for x in raw_variables if x != None]
        if len(variables) < len(raw_variables):
            print(
                colored(
                    f"There are some blank columns inside the table sheet {self.sheet_title}, you may delete those blank columns at the right of your visiable columns",
                    "red",
                )
            )
        return variables

    # TODO: convert?
    def _title2variable(self, title):
        return title.lower().replace(" ", "_")

    # TODO: ?
    def _variable2title(self, variable):
        return variable.replace("_", " ").title()

    def __str__(self):
        return self.sheet_title

    # hash only care about variables and values
    def __hash__(self) -> int:
        return hash(",".join([node for node in self.data]))

    def __len__(self):
        return len(self.data)

    # TODO: has same value. __eq__ only cares about has same schema, instead of values. So, has same value will compare value
    def has_same_value(self, another):
        pass

    # == > < All this operations do not consider value. They only compare sheet variables
    # == return True if self and another has same variables. Only care about variables
    def __eq__(self, another: object) -> bool:
        # check if sheet has same column variables, which means same data structure
        if another == None:
            return False
        if self.data == [] and another == []:
            return True
        if self.column_variables != another.column_variables:
            return False
        return True

    # # > return True if every another object's items is in self, and the length of self is greater than the other's
    def __gt__(self, another: object) -> bool:
        flags = []
        for variable in another.column_variables:
            flags.append(True) if variable in self.column_variables else flags.append(
                False
            )
        flags.append(len(self.column_variables) > len(another.column_variables))
        return all(flags)

    # + union two objects (self | another)'s two lists(column_titles, column_variables), and return a new object
    def __add__(self, another: object):
        obj = copy.deepcopy(
            self
        )  # deep copy self to obj, to avoid messing up self object
        # get titles united. list the OrderedDict union result for remove duplication
        # get variables
        obj.column_variables = list(
            OrderedDict.fromkeys(self.column_variables + another.column_variables)
        )  # 字典keys相加去重复
        titles = [
            self.column_titles.get(v) or another.column_titles.get(v)
            for v in obj.column_variables
        ]  # get new set of titles
        obj.column_titles = OrderedDict(
            zip(obj.column_variables, titles)
        )  # re-bind to new set of variables
        another.column_titles, another.column_variables = (
            obj.column_titles,
            obj.column_variables,
        )  # 让another和obj一样的titles和variables

        # expand obj and another to the new union. As a result, they will be in same number of columns
        for o in obj.data:
            for p in obj.column_variables:
                if not getattr(
                    o, p, None
                ):  # if the property has no value, set the property as None.
                    setattr(o, p, None)

        for o in another.data:
            for p in another.column_variables:  # here obj.column_variables are expaned.
                if not getattr(o, p, None):
                    setattr(o, p, None)

        # Now self and another are in same dimension in columns。
        temp_data = []
        for node_obj in obj.data:  # 对每一个obj里面的数据 去寻找是否和another里面的是相同记录
            for node_another in another.data:
                if node_obj.is_same_record(node_another):  # if same record
                    node_obj += another.data.pop(
                        another.data.index(node_another)
                    )  # pop up the same one and merge
                    temp_data.append(node_obj)  # 将merge过的记录添加到临时list中
                else:
                    if node_obj not in temp_data:  # 如果不是同一条记录，而且在临时列表中没有，添加，
                        temp_data.append(node_obj)

        obj.data = temp_data + another.data  # append all different rows to obj.data。
        # 此事obj。data数据都有了，但是order有可能发生改变，所以下面需要对obj的data进行重构，以variables的顺序作为统一标准。
        # 根据obj.data中node的variable排序，重新构建obj。
        new_data = []
        for d in obj.data:
            values = [
                getattr(d, p, None) for p in obj.column_variables
            ]  # 以variable取出对应的value
            pair = OrderedDict(
                zip(obj.column_variables, values)
            )  # zip variable和value，编程有序字典
            new_data.append(TableNode(**pair))  # 重新构建TableNode obj到临时data列表
        obj.data = new_data
        return obj

    # # - operator: 1st: minus only allowed if self and another have same column schema(same variables). 2st, if top 3 variables are same, the row will be removed
    # def __sub__(self, another: object):
    #     obj = copy.deepcopy(
    #         self
    #     )  # deep copy self to obj, to avoid messing up self object
    #     # check if has same column schema, if not raise an error
    #     if self.column_variables != another.column_variables:
    #         raise ValueError(
    #             f"{self.column_titles} is not same as {another.column_variables}, so could not minus"
    #         )
    #     # check for same row, if yes, remove it
    #     for obj_node in self.data:  # use self.data because obj.data is changing...
    #         for another_node in another.data:
    #             if obj_node.is_same_record(another_node):
    #                 obj.data.pop(obj.data.index(obj_node))

    #     return obj

    # - set difference, return a new set with elements in self but not in another
    def __sub__(self, another: object):
        # deep copy self to obj, to avoid messing up self object
        obj = copy.deepcopy(self)

        if self.column_variables != another.column_variables:
            raise ValueError(
                f"Excel table-sheet minus is only used for same table schema, and remove row in self which also in another.  {self.column_titles} is not same as {another.column_variables}, so could not minus"
            )

        for obj_node in self.data:  # use self.data because obj.data is changing...
            # remove node in self, if it is also in another
            variables = obj_node.__dict__.keys()
            if obj_node in another.data and (
                "variable_type" not in variables or "display_type" not in variables
            ):
                obj.data.remove(obj_node)
        return obj

    # in: return True if self's variables contains another's
    def __contain__(self, another: object):
        for x in another.column_variables:
            if x not in self.column_variables:
                return False
        return True

    def createTableNodes(self, rows):
        for row in rows:
            row_data = OrderedDict(
                zip(self.column_variables, row)
            )  # zip and make data as a dict
            self.data.append(TableNode(**row_data))  # create a InfoNode object

    def showTableNodes(self, seperator=","):
        for row in self.data:
            info = [info for info in row.__dict__.values()]
            info = [str(i) if i is not None else "" for i in info]  # avoid None
            print(seperator.join(info))

    def copy(self, another: object):
        obj = copy.deepcopy(self)
        # for some tables, has to be specially processed
        if another.name in ["personid", "address", "phone", "eraddress", "contact"]:
            # loop another.data
            for ad in another.data:
                # loop obj.data
                for index, od in enumerate(obj.data):
                    # if another.data's variable ==obj.data's variable_type, then copy
                    if (
                        ad.variable_type == od.variable_type
                        and ad.display_type == od.display_type
                    ):  # using display_type is only for working_address, which has multiple value in er.xlsx
                        obj.data[index] = od.copy(ad)
                        break
            return obj
        # for general tables
        for ad in another.data:
            # if has same schema
            if self.column_variables == another.column_variables:
                # if another not in original one, just append another's rows to original one.
                if ad not in obj.data:
                    obj.data.append(ad)
            # if has different schema
            else:
                # get self's column variable,and get value from another, then make a TableNode object and insert to self.data
                kwargs = {}
                for key in obj.column_variables:
                    kwargs[key] = getattr(ad, key, None)
                copied_obj = TableNode(**kwargs)
                obj.data.append(copied_obj)
        return obj

    # get common variables in self and another, and return new object
    def common(self, another: object):
        obj = copy.deepcopy(self)
        # get titles united. list the OrderedDict union result for remove duplication
        obj.column_titles = [
            {k: v}
            for k, v in self.column_titles.items()
            if k in another.column_titles.keys()
        ]
        # get variables
        obj.column_variables = [
            x for x in self.column_variables if x in another.column_variables
        ]
        # get values
        obj.data = [x for x in self.data if x in another.data]
        return obj

    # Iterator
    def __iter__(self):
        return self

    def __next__(self):
        try:
            item = self.data[self.index]
        except IndexError:
            raise StopIteration()
        self.index += 1
        return item



from pydantic import BaseModel, root_validator,validator,EmailStr
from datetime import date
from typing import Optional
from model.common.utils import makeList,checkRow
from model.common.person import Person
from model.common.mixins import DurationMixin
from model.common.employmentbase import EmploymentBase
from model.common.educationbase import EducationBase

class PersonalAssess(BaseModel):
    self_description:Optional[str]
    skill_list:list
    activity:Optional[list]
    
    _str2bool_activity=validator('activity',allow_reuse=True,pre=True)(makeList)
    _str2bool_skill_list=validator('skill_list',allow_reuse=True,pre=True)(makeList)

class Personal(Person):
    email:Optional[EmailStr]
    
    @property
    def birth_day(self):
        return self.dob.strftime('%b %d, %Y')
    
class Education(EducationBase):
    city:str
    country:str
    description:Optional[str]
    
    @root_validator
    def checkRowCompletion(cls,values):
        all_fields=[
            "start_date",
            "end_date",
            "school_name",
            "education_level",
            "field_of_study",
            "city",
            "country",
            "description"
            ]
        required_fields=[
            "start_date",
            "school_name",
            "education_level",
            "field_of_study",
            "city",
            "country"
        ]

        checkRow(values, all_fields, required_fields)
        
        return values

    

class Language(BaseModel):
    reading:str
    writting:str
    listening:str
    speaking:str
    test_type: str
    remark: Optional[str]
    
    @root_validator
    def checkRowCompletion(cls,values):

        all_fields=["reading",'writting','listening','speaking','test_type',"remark"]
        required_fields=["reading",'writting','listening','speaking','test_type']
        checkRow(values, all_fields, required_fields)
        
        return values
    
class Employment(EmploymentBase):
    duties:list
    
    _str2bool_duties=validator('duties',allow_reuse=True,pre=True)(makeList)
    
    @root_validator
    def checkCanadaProvince(cls, values):
        country = values["country"]
        province = values["province"]
        if (
            country.lower() == "canada"
            and province
            not in [
                "AB",
                "BC",
                "MB",
                "NB",
                "NL",
                "NS",
                "NT",
                "NU",
                "ON",
                "PE",
                "QC",
                "SK",
                "YT",
            ]
        ):
            raise ValueError(
                f'Since country is Canada is, the province must be one of  "AB","BC","MB","NB","NL","NS","NT","NU","ON","PE","QC","SK","YT"'
            )
        return values
    
    @root_validator
    def checkRowCompletion(cls,values):
        all_fields=[
                    "start_date",
                    "end_date",
                    "job_title",
                    "noc_code",
                    "weekly_hours",
                    "company",
                    "city",
                    "province",
                    "country",
                    "duties"
                ]
        all_fields_values=[values.get(field,None) for field in all_fields]
        
        required_fields=[
                    "start_date",
                    "job_title",
                    "noc_code",
                    "weekly_hours",
                    "company",
                    "city",
                    "country",
                    "duties"
                ]
        required_values=[values.get(field,None) for field in required_fields]
        
        has_values=[value for value in required_values if value]
        
        if any(all_fields_values) and not all(required_values):
            raise ValueError(f"Please check the row with values ({has_values}) in table-address, some required fileds are missed.")
        return values
    



from model.common.commonmodel import CommonModel
from datetime import date
from pydantic import BaseModel, root_validator
from typing import List
from model.common.mixins import DatePeriod
from model.common.utils import checkRow


class Travel(DatePeriod):
    length: int
    destination: str
    purpose: str

    @root_validator
    def checkCompletion(cls, values):
        all_fields = ["start_date", "end_date", "length", "destination", "purpose"]
        required_fields = ["start_date", "end_date", "length", "destination", "purpose"]
        checkRow(values, all_fields, required_fields)
        return values


class Personal(BaseModel):
    last_name: str
    first_name: str


class M5562Model(CommonModel):
    travel: List[Travel]
    personal: Personal

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        if output_excel_file:
            excels = self.getExcels(["excel/pa.xlsx"])
        else:
            if excels is None and len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        super().__init__(excels, output_excel_file, globals())



from datetime import date
from pydantic import BaseModel, EmailStr, root_validator
from typing import List, Union
from model.common.mixins import DatePeriod
from model.common.utils import checkRow


class COR(DatePeriod):
    start_date: date
    end_date: Union[date, None]
    country: str
    status: str

    @root_validator
    def checkCompletion(cls, values):
        all_fields = [
            "country",
            "status",
            "start_date",
            "end_date",
        ]
        required_fields = [
            "country",
            "status",
            "start_date",
            "end_date",
        ]
        checkRow(values, all_fields, required_fields)

        return values


class CORs(object):
    # Frist row must be current residence
    def __init__(self, cors: List[COR]):
        self.cors = cors

    @property
    def current(self):
        ccor = [country for country in self.cors if country.end_date == None]
        if len(ccor) == 1:
            return ccor[0]
        elif len(ccor) == 0 and len(self.cors) > 0:
            return self.cors[0]
        else:
            raise ValueError("No residence data, please check")

    @property
    def previous(self):
        if len(self.cors) == 0:
            print("There is no residence data, please check")
            return
        if len(self.cors) == 1:
            print(
                "There is only one row data, which is considered current residence. So, there is no previous residence."
            )
            return
        return self.cors[1:]



from termcolor import colored
from utils.utils import checkContinuity
from typing import List, Optional, Union
from model.common.address import Address
from model.common.educationbase import EducationBase
from model.common.phone import Phone
from model.common.commonmodel import CommonModel
from model.common.id import ID
from model.common.person import Person
from datetime import date
from pydantic import BaseModel, EmailStr, root_validator, validator
from model.common.utils import normalize
from webform.prportal.data.country import country_residence
from model.common.mixins import DatePeriod


class PersonId(ID):
    pass


class PRCase(BaseModel):
    imm_program: str
    imm_category: str
    imm_under: Optional[str]
    communication_language: str
    interview_language: str
    need_translator: bool
    intended_province: str
    intended_city: str


class COR(DatePeriod):
    country: str
    status: str
    explanation: Optional[str]


class CORs(object):
    def __init__(self, cors: List[COR]):
        self.cors = cors
        self.checkFirstEndDate()

    @property
    def current(self):
        if len(self.cors) >= 1:
            return self.cors[0]
        else:
            raise ValueError(
                colored(
                    f"table-cor must have values, and the first line must be current residence country.",
                    "red",
                )
            )

    # In table cor, the first line is defined as current residence, so the end date must bigger than today
    def checkFirstEndDate(self):
        if (
            len(self.cors) > 0
            and self.cors[0].end_date
            and self.cors[0].end_date < date.today()
        ):
            raise ValueError(
                colored(
                    f"In tabel-cor, the current residence end date {self.cors[0].end_date} is earlier than today",
                    "red",
                )
            )


class Marriage(BaseModel):
    marital_status: str
    married_date: Optional[date]
    sp_last_name: Optional[str]
    sp_first_name: Optional[str]
    previous_married: bool
    pre_sp_last_name: Optional[str]
    pre_sp_first_name: Optional[str]
    pre_sp_dob: Optional[date]
    pre_relationship_type: Optional[str]
    pre_start_date: Optional[date]
    pre_end_date: Optional[date]

    @validator("pre_end_date")
    def endDateBigger(cls, pre_end_date, values):
        pre_start_date = values.get("pre_start_date")
        if not pre_start_date:
            return pre_end_date
        the_date = pre_end_date or date.today()
        if (the_date - pre_start_date).days <= 0:
            raise ValueError(
                f"Expiry date {the_date} is earlier than issue date {pre_start_date} for the ID"
                if pre_end_date
                else f"{pre_start_date} is later than today"
            )
        return pre_end_date

    @root_validator
    def checkPreviousMarriage(cls, values):
        pre_marriaged = values["previous_married"]
        pre_sp_last_name = values["pre_sp_last_name"]
        pre_sp_first_name = values["pre_sp_first_name"]
        pre_sp_dob = values["pre_sp_dob"]
        pre_start_date = values["pre_start_date"]
        pre_end_date = values["pre_end_date"]
        pre_relationship_type = values["pre_relationship_type"]

        if pre_marriaged and not all(
            [
                pre_sp_first_name,
                pre_sp_last_name,
                pre_sp_dob,
                pre_start_date,
                pre_end_date,
                pre_relationship_type,
            ]
        ):
            raise ValueError(
                f"Since previous married is true, but you did not answer all the questions for previous spouse: first/last name, dob,relationship type, start/end date"
            )
        return values

    @root_validator
    def checkCurrentMarriage(cls, values):
        marital_status = values["marital_status"]
        married_date = values["married_date"]
        sp_last_name = values["sp_last_name"]
        sp_first_name = values["sp_first_name"]

        if marital_status in ["Common-Law", "Married"] and not all(
            [married_date, sp_last_name, sp_first_name]
        ):
            raise ValueError(
                f"Marital stauts is {marital_status}, but you did not answer all the questions for spouse: first/last name, or marriage date"
            )
        return values


class Personal(Person):
    used_last_name: Optional[str]
    used_first_name: Optional[str]
    native_last_name: str
    native_first_name: str
    height: int
    eye_color: str
    country_of_birth: str
    place_of_birth: str
    uci: Optional[str]
    citizen2: Optional[str]
    native_language: str
    english_french: str
    which_one_better: Optional[str]
    language_test: bool
    current_occupation: str
    intended_occupation: Optional[str]
    email: EmailStr
    primary_school_years: int
    secondary_school_years: int
    post_secondary_school_years: int
    other_school_years: int
    other_explanation: Optional[str]
    relationship_to_pa: Optional[str]
    accompany_to_canada: Optional[bool]
    dependant_type: Optional[str]

    _normalize_used_first_name = validator(
        "used_first_name", allow_reuse=True, check_fields=False
    )(normalize)
    _normalize_used_last_name = validator(
        "used_last_name", allow_reuse=True, check_fields=False
    )(normalize)

    @root_validator
    def checkAnswers(cls, values):
        english_french = values["english_french"]
        explanations = values["which_one_better"]
        if english_french == "Both" and not explanations:
            raise ValueError(
                colored(
                    f"Since you answered you can speak both English and French, but you did not answer the question 'which on is better' in info-personal sheet",
                    "red",
                )
            )
        return values

    @root_validator
    def checkRelationship(cls, values):
        relationship = values["relationship_to_pa"]
        accompany_to_canada = values["accompany_to_canada"]
        dependant_type = values["dependant_type"]

        if relationship and (accompany_to_canada == None):
            raise ValueError(
                colored(
                    f"In info-personal, since {values['first_name']} {values['last_name']} is {relationship} to principle applicant, but you did not pick if accompany to Canada.",
                    "red",
                )
            )

        if (
            relationship
            and relationship
            in ["Adopted Child", "Child", "Grandchild", "Step-Child", "Step-Grandchild"]
            and dependant_type == None
        ):
            raise ValueError(
                colored(
                    f"In info-personal, since {values['first_name']} {values['last_name']} is {relationship} to principle applicant, but you did not pick dependant type",
                    "red",
                )
            )
        return values


class Education(EducationBase):
    city: Optional[str]
    country: Optional[str]


class Family(BaseModel):
    relationship: Optional[str]
    last_name: Optional[str]
    first_name: Optional[str]
    native_last_name: Optional[str]
    native_first_name: Optional[str]
    date_of_birth: Optional[date]
    date_of_death: Optional[date]
    place_of_birth: Optional[str]
    birth_country: Optional[str]
    marital_status: Optional[str]
    email: Optional[EmailStr]
    address: Optional[str]

    @root_validator
    def checkFields(cls, values):
        last_name = values["last_name"]
        first_name = values["first_name"]
        native_last_name = values["native_last_name"]
        native_first_name = values["native_first_name"]
        date_of_birth = values["date_of_birth"]
        place_of_birth = values["place_of_birth"]
        birth_country = values["birth_country"]
        marital_status = values["marital_status"]
        address = values["address"]

        if (
            last_name
            and first_name
            and not all(
                [
                    native_last_name,
                    native_first_name,
                    date_of_birth,
                    place_of_birth,
                    birth_country,
                    marital_status,
                    address,
                ]
            )
        ):
            raise ValueError(
                f"Some fields of {first_name} {last_name} in info-family sheet are not completed, please check"
            )
        return values


class Travel(DatePeriod):
    length: int
    destination: str
    purpose: str


class PRBackground(BaseModel):
    q1: bool
    q2: bool
    q3: bool
    q4: bool
    q5: bool
    q6: bool
    q7: bool
    q8: bool
    q9: bool
    q10: bool
    q11: bool
    details: Optional[str]


class History(DatePeriod):
    activity: str
    city_and_country: str
    status: str
    name_of_company_or_school: str

    @property
    def __str__(self):
        return "table-history"


class Member(DatePeriod):
    organization_name: str
    organization_type: str
    position: str
    city: str
    country: str


class Government(DatePeriod):
    country: str
    department: str
    position: str


class Military(DatePeriod):
    country: str
    service_detail: str
    rank: str
    combat_detail: Optional[str]
    reason_for_end: Optional[str]


class AddressHistory(DatePeriod):
    street_and_number: str
    city: str
    province: str
    country: str
    post_code: str

    @property
    def __str__(self):
        return "table-addresshistory"


class Status(BaseModel):
    last_entry_date: Optional[date]
    last_entry_place: Optional[str]


class PrModel(CommonModel):
    personal: Personal
    status: Status
    cor: List[COR]
    marriage: Marriage
    address: List[Address]
    phone: List[Phone]
    personid: List[PersonId]
    prcase: PRCase
    family: List[Family]
    travel: List[Travel]
    prbackground: PRBackground
    education: List[Education]
    history: List[History]
    member: List[Member]
    government: List[Government]
    military: List[Military]
    addresshistory: List[AddressHistory]

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None, check=True):
        # check continuity only if check is required and with excels
        should_check = check and excels
        if output_excel_file:
            excels = self.getExcels(["excel/pr.xlsx", "excel/pa.xlsx"])
        else:
            if excels is None and len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        super().__init__(excels, output_excel_file, globals())

        if should_check:
            [
                self.check(item, excels[0])
                for item in [self.history, self.addresshistory]
            ]

    def check(self, items: List[object], excel_file):
        results = []
        # construct list suitable for checkContinuity
        for item in items:
            results.append(list(item.__dict__.values()))

        # check
        continued, sorted_list, msg = checkContinuity(results)

        if not continued:
            print(
                colored(
                    f"There are {len(msg)} error(s) in sheet {items[0].__str__} of {excel_file}",
                    "red",
                )
            )
            [print(index, "\t", m) for index, m in enumerate(msg)]



data_model = {
    "trcasein": {
        "service_in": "Page1/PersonalDetails/ServiceIn/ServiceIn",
        "original_entry_date": "Page3/ComingIntoCda/OrigEntry/DateLastEntry",
        "original_entry_place": "Page3/ComingIntoCda/OrigEntry/Place",
        "original_purpose": "Page3/ComingIntoCda/PurposeOfVisit/PurposeOfVisit",
        "original_other_reason": "Page3/ComingIntoCda/PurposeOfVisit/Other",
        "most_recent_entry_date": "Page3/ComingIntoCda/RecentEntry/DateLastEntry",
        "most_recent_entry_place": "Page3/ComingIntoCda/RecentEntry/Place",
        "doc_number": "Page3/ComingIntoCda/PrevDocNum/docNum",
        "is_spouse_canadian": "Page1/MaritalStatus/d/SpouseStatus",
        "consent_of_info_release": "Page4/Signature/FutureComm",
        "submission_date": "Page4/Signature/C1CertificateIssueDate",
    },
    "wpincanada": {
        "application_purpose1": "Page1/PersonalDetails/ApplyingFor/RestoreStat",
        "application_purpose2": "Page1/PersonalDetails/ApplyingFor/Extend",
        "application_purpose3": "Page1/PersonalDetails/ApplyingFor/TRP",
        "caq_number": "Page3/DetailsOfWork/CAQ/CertNum",
        "expiry_date": "Page3/DetailsOfWork/CAQ/CertExpiry",
        "work_permit_type": "Page3/DetailsOfWork/Purpose/Type",
        "employer_name": "Page3/DetailsOfWork/Employer/Name",
        "employer_address": "Page3/DetailsOfWork/Employer/Addr",
        "work_province": "Page3/DetailsOfWork/Location/Prov",
        "work_city": "Page3/DetailsOfWork/Location/City",
        "work_address": "Page3/DetailsOfWork/Location/Addr",
        "job_title": "Page3/DetailsOfWork/Occupation/Job",
        "brief_duties": "Page3/DetailsOfWork/Occupation/Desc",
        "start_date": "Page3/DetailsOfWork/Duration/FromDate",
        "end_date": "Page3/DetailsOfWork/Duration/ToDate",
        "lmia_num_or_offer_num": "Page3/DetailsOfWork/Duration/LMO",
    },
    "personal": {
        "uci": "Page1/PersonalDetails/ServiceIn/UCIClientID",
        "last_name": "Page1/PersonalDetails/Name/FamilyName",
        "first_name": "Page1/PersonalDetails/Name/GivenName",
        "has_alias_name": "Page1/PersonalDetails/AliasName/AliasNameIndicator/AliasNameIndicator",
        "used_last_name": "Page1/PersonalDetails/AliasName/AliasFamilyName",
        "used_first_name": "Page1/PersonalDetails/AliasName/AliasGivenName",
        "sex": "Page1/PersonalDetails/q3-4-5/sex/Sex",
        "dob_year": "Page1/PersonalDetails/q3-4-5/dob/DOBYear",
        "dob_month": "Page1/PersonalDetails/q3-4-5/dob/DOBMonth",
        "dob_day": "Page1/PersonalDetails/q3-4-5/dob/DOBDay",
        "place_of_birth": "Page1/PersonalDetails/q3-4-5/pob/PlaceBirthCity",
        "country_of_birth": "Page1/PersonalDetails/q3-4-5/pob/PlaceBirthCountry",
        "citizen": "Page1/PersonalDetails/Citizenship/Citizenship",
        "native_language": "Page2/Languages/nativeLang",
        "english_french": "Page2/Languages/communicateLang",
        "which_one_better": "Page2/Languages/FreqLang",
        "language_test": "Page2/Languages/LangTestIndicator",
        "email": "Page2/ContactInformation/q5-6/Email/Email",
    },
    # country of residence: current and past for over 6 months. table model
    "cor": {
        "current_cor_country": "Page1/PersonalDetails/CurrentCOR/CurrentCOR/Row2/Country",
        "current_cor_status": "Page1/PersonalDetails/CurrentCOR/CurrentCOR/Row2/Status",
        "current_cor_other": "Page1/PersonalDetails/CurrentCOR/CurrentCOR/Row2/Other",
        "current_cor_start_date": "Page1/PersonalDetails/CurrentCOR/CurrentCOR/Row2/FromDate",
        "current_cor_end_date": "Page1/PersonalDetails/CurrentCOR/CurrentCOR/Row2/ToDate",
        "previous_cor_country1": "Page1/PersonalDetails/PrevCOR/PreviousCOR/Row2/Country",
        "previous_cor_status1": "Page1/PersonalDetails/PrevCOR/PreviousCOR/Row2/Status",
        "previous_cor_other1": "Page1/PersonalDetails/PrevCOR/PreviousCOR/Row2/Other",
        "previous_cor_start_date1": "Page1/PersonalDetails/PrevCOR/PreviousCOR/Row2/FromDate",
        "previous_cor_end_date1": "Page1/PersonalDetails/PrevCOR/PreviousCOR/Row2/ToDate",
        "previous_cor_country2": "Page1/PersonalDetails/PrevCOR/PreviousCOR/Row3/Country",
        "previous_cor_status2": "Page1/PersonalDetails/PrevCOR/PreviousCOR/Row3/Status",
        "previous_cor_other2": "Page1/PersonalDetails/PrevCOR/PreviousCOR/Row3/Other",
        "previous_cor_start_date2": "Page1/PersonalDetails/PrevCOR/PreviousCOR/Row3/FromDate",
        "previous_cor_end_date2": "Page1/PersonalDetails/PrevCOR/PreviousCOR/Row3/ToDate",
    },
    # marriage
    "marriage": {
        "marital_status": "Page1/MaritalStatus/Current/MaritalStatus",
        "married_date": "Page1/MaritalStatus/Current/b/DateOfMarriage",
        "sp_last_name": "Page1/MaritalStatus/Current/c/FamilyName",
        "sp_first_name": "Page1/MaritalStatus/Current/c/GivenName",
        "previous_married": "Page2/MaritalStatus/PrevMarriage/PrevMarriedIndicator",
        "pre_sp_last_name": "Page2/MaritalStatus/PrevMarriage/PMFamilyName",
        "pre_sp_first_name": "Page2/MaritalStatus/PrevMarriage/PMGivenName",
        "pre_relationship_type": "Page2/MaritalStatus/PrevMarriage/TypeOfRelationship",
        "pre_sp_dob_year": "Page2/MaritalStatus/PrevMarriage/dob/DOBYear",
        "pre_sp_dob_month": "Page2/MaritalStatus/PrevMarriage/dob/DOBMonth",
        "pre_sp_dob_day": "Page2/MaritalStatus/PrevMarriage/dob/DOBDay",
        "pre_start_date": "Page2/MaritalStatus/PrevMarriage/From/FromDate",
        "pre_end_date": "Page2/MaritalStatus/PrevMarriage/To/ToDate",
    },
    "passport": {
        "number": "Page2/Passport/PassportNum",
        "country": "Page2/Passport/CountryofIssue",
        "issue_date": "Page2/Passport/IssueDate",
        "expiry_date": "Page2/Passport/ExpiryDate",
    },
    "national_id": {
        "has_id": "Page2/natID/q1/natIDIndicator",
        "number": "Page2/natID/natIDdocs/DocNum/DocNum",
        "country": "Page2/natID/natIDdocs/CountryofIssue/CountryofIssue",
        "issue_date": "Page2/natID/natIDdocs/IssueDate/IssueDate",
        "expiry_date": "Page2/natID/natIDdocs/ExpiryDate",
    },
    "us_pr": {
        "has_id": "Page2/USCard/q1/usCardIndicator",
        "number": "Page2/USCard/usCarddocs/DocNum/DocNum",
        "expiry_date": "Page2/USCard/usCarddocs/ExpiryDate",
    },
    "mailing_address": {
        "po_box": "Page2/ContactInformation/Mailing/AddrLine1/POBox",
        "unit": "Page2/ContactInformation/Mailing/AddrLine1/AptUnit",
        "street_number": "Page2/ContactInformation/Mailing/AddrLine1/StreetNum",
        "street_name": "Page2/ContactInformation/Mailing/AddrLine1/Streetname",
        "city": "Page2/ContactInformation/Mailing/AddrLine2/City",
        "country": "Page2/ContactInformation/Mailing/AddrLine2/Country",
        "province": "Page2/ContactInformation/Mailing/AddrLine2/Prov",
        "post_code": "Page2/ContactInformation/Mailing/AddrLine2/PostalCode",
        "district": "Page2/ContactInformation/Mailing/AddrLine2/District",
    },
    "residential_address": {
        "same_as_mailing": "Page2/ContactInformation/Resi/SameAsAddr/SameAsMailingInd",
        "unit": "Page2/ContactInformation/Resi/AddrLine1/AptUnit",
        "street_number": "Page2/ContactInformation/Resi/AddrLine1/StreetNum",
        "street_name": "Page2/ContactInformation/Resi/AddrLine1/Streetname",
        "city": "Page2/ContactInformation/Resi/AddrLine2/City",
        "country": "Page2/ContactInformation/Resi/AddrLine2/Country",
        "province": "Page2/ContactInformation/Resi/AddrLine2/Prov",
        "post_code": "Page2/ContactInformation/Resi/AddrLine2/PostalCode",
        "district": "Page2/ContactInformation/Resi/AddrLine2/District",
    },
    "phone": {
        "variable_type": "Page2/ContactInformation/q3-4/Phone/Type",
        "canada_us": "Page2/ContactInformation/q3-4/Phone/CanOtherInd/CanadaUS",
        "other": "Page2/ContactInformation/q3-4/Phone/CanOtherInd/Other",
        "ext": "Page2/ContactInformation/q3-4/Phone/NumberExt",
        "country_code": "Page2/ContactInformation/q3-4/Phone/NumberCountry",
        "number": "Page2/ContactInformation/q3-4/Phone/ActualNumber",
        "na_area": "Page2/ContactInformation/q3-4/Phone/NANumber/AreaCode",
        "na_first_three": "Page2/ContactInformation/q3-4/Phone/NANumber/FirstThree",
        "na_last_five": "Page2/ContactInformation/q3-4/Phone/NANumber/LastFive",
        "int_number": "Page2/ContactInformation/q3-4/Phone/IntlNumber/IntlNumber",
    },
    "altphone": {
        "variable_type": "Page2/ContactInformation/q3-4/AltPhone/Type",
        "canada_us": "Page2/ContactInformation/q3-4/AltPhone/CanOtherInd/CanadaUS",
        "other": "Page2/ContactInformation/q3-4/AltPhone/CanOtherInd/Other",
        "ext": "Page2/ContactInformation/q3-4/AltPhone/NumberExt",
        "country_code": "Page2/ContactInformation/q3-4/AltPhone/NumberCountry",
        "number": "Page2/ContactInformation/q3-4/AltPhone/ActualNumber",
        "na_area": "Page2/ContactInformation/q3-4/AltPhone/NANumber/AreaCode",
        "na_first_three": "Page2/ContactInformation/q3-4/AltPhone/NANumber/FirstThree",
        "na_last_five": "Page2/ContactInformation/q3-4/AltPhone/NANumber/LastFive",
        "int_number": "Page2/ContactInformation/q3-4/AltPhone/IntlNumber/IntlNumber",
    },
    "fax": {
        "canada_us": "Page2/ContactInformation/q5-6/Fax/CanOtherInd/CanadaUS",
        "other": "Page2/ContactInformation/q5-6/Fax/CanOtherInd/Other",
        "ext": "Page2/ContactInformation/q5-6/Fax/NumberExt",
        "country_code": "Page2/ContactInformation/q5-6/Fax/NumberCountry",
        "number": "Page2/ContactInformation/q5-6/Fax/ActualNumber",
        "na_area": "Page2/ContactInformation/q5-6/Fax/NANumber/AreaCode",
        "na_first_three": "Page2/ContactInformation/q5-6/Fax/NANumber/FirstThree",
        "na_last_five": "Page2/ContactInformation/q5-6/Fax/NANumber/LastFive",
        "int_number": "Page2/ContactInformation/q5-6/Fax/IntlNumber/IntlNumber",
    },
    "education": {
        "start_date_year": "Page3/Education/EduLine1/From/YYYY",
        "start_date_month": "Page3/Education/EduLine1/From/MM",
        "end_date_year": "Page3/Education/EduLine2/To/YYYY",
        "end_date_month": "Page3/Education/EduLine2/To/MM",
        "school_name": "Page3/Education/EduLine1/School",
        "field_of_study": "Page3/Education/EduLine1/FieldOfStudy",
        "city": "Page3/Education/EduLine2/City",
        "country": "Page3/Education/EduLine2/Country",
        "province": "Page3/Education/EduLine2/Prov",
    },
    # occupation is a special dict, and will be handled seperately
    "occupation": {
        "start_date_year": "Line1/From/YYYY",
        "start_date_month": "Line1/From/MM",
        "end_date_year": "Line2/To/YYYY",
        "end_date_month": "Line2/To/MM",
        "job_title": "Line1/Occupation",
        "company": "Line1/Employer",
        "city": "Line2/City",
        "province": "Line2/ProvState",
        "country": "Line2/Country",
    },
    "trbackground": {
        "q1a": "Page4/BackgroundInfo/HealthQ/qANY",
        "q1b": "Page4/BackgroundInfo/HealthQ/qBNY",
        "q1c": "Page4/BackgroundInfo/HealthQ/MedicalDetails",
        "q2a": "Page4/BackgroundInfo/PrevApplied/qANY",
        "q2b": "Page4/BackgroundInfo/PrevApplied/qBNY",
        "q2c": "Page4/BackgroundInfo/PrevApplied/qCNY",
        "q2d": "Page4/BackgroundInfo/PrevApplied/refusedDetails",
        "q3a": "Page4/BackgroundInfo/Criminal/qANY",
        "q3b": "Page4/BackgroundInfo/Criminal/refusedDetails",
        "q4a": "Page4/BackgroundInfo/Military/qANY",
        "q4b": "Page4/BackgroundInfo/Military/militaryServiceDetails",
        "q5": "Page4/BackgroundInfo/GovPosition/qGovtNY",
        "q6": "Page4/BackgroundInfo/Illtreatment/qWitnessNY",
    },
}
convert_model = {
    "country_pairs": {
        "trcase": ["applying_country"],
        "personal": ["country_of_birth", "citizen"],
        "passport": ["country"],
        "national_id": ["country"],
        "education": ["country"],
        "cor": [
            "current_cor_country",
            "previous_cor_country1",
            "previous_cor_country2",
        ],
    },
    "canada_province_pairs": {
        "wpincanada": ["work_province"],
        "address": ["province"],
        "education": ["province"],
        "employment": ["province"],
    },
    "language_pairs": {"personal": ["native_language"]},
    "phone_type_pairs": {"phone": ["variable_type"], "altphone": ["variable_type"]},
    "tr_portal_study_level_pairs": {"sp": ["study_level"]},
    "tr_portal_study_field_pairs": {"sp": ["study_field"]},
    "tr_canada_status_pairs": {
        "cor": ["current_cor_status", "previous_cor_status1", "previous_cor_status2"]
    },
    "tr_marital_status_pairs": {
        "marriage": ["marital_status", "pre_relationship_type"]
    },
}

remove_model = {
    "personal": ["dob_year", "dob_month", "dob_day"],
    "marriage": ["pre_sp_dob_year", "pre_sp_dob_month", "pre_sp_dob_day"],
}



from datetime import date
from pydantic import BaseModel, root_validator, validator
from typing import Optional, List
from model.common.address import Address

# 1. start date must be greater than today
# 2. end date must be greater than start date
class DurationModel(BaseModel):
    start_date: date
    end_date: date

    @validator("end_date")
    def endDateBigger(cls, end_date, values):
        start_date = values.get("start_date")
        the_date = end_date or date.today()
        if (the_date - start_date).days <= 0:
            raise ValueError(
                f"End date {the_date} is earlier than from date {start_date} for the work period"
            )
        return end_date

    @validator("start_date")
    def checkStartDate(cls, v):
        if (v - date.today()).days <= 0:
            raise ValueError("The start date of work must be greater than today")
        return v


class TrCase(BaseModel):
    service_in: str
    same_as_cor: bool
    applying_country: Optional[str]
    applying_status: Optional[str]
    applying_start_date: Optional[date]
    applying_end_date: Optional[date]
    consent_of_info_release: bool
    submission_date: Optional[date]


class TrCaseIn(BaseModel):
    service_in: str
    application_purpose: Optional[str]  # TODO: consider
    original_entry_date: date
    original_entry_place: str
    original_purpose: str
    original_other_reason: Optional[str]
    most_recent_entry_date: Optional[date]
    most_recent_entry_place: Optional[str]
    doc_number: Optional[str]
    consent_of_info_release: bool
    submission_date: Optional[date]

    @root_validator
    def checkMostRecent(cls, values):
        most_recent_entry_date = values.get("most_recent_entry_date")
        most_recent_entry_place = values.get("most_recent_entry_place")
        original_entry_date = values.get("original_entry_date")
        if (
            most_recent_entry_date
            and original_entry_date
            and (most_recent_entry_date - original_entry_date).days <= 0
        ):
            raise ValueError(
                f"The most recent visiting date must later than the original one, but your input {most_recent_entry_date} is equal/earlier to the original {original_entry_date}"
            )

        if most_recent_entry_date and not most_recent_entry_place:
            raise ValueError(
                "You inputed most recent entry date, but most recent entry place is missed."
            )
        return values


class Visa(DurationModel):
    application_purpose: str
    visit_purpose: str
    funds_available: int
    name1: Optional[str]
    relationship1: Optional[str]
    address1: Optional[str]
    name2: Optional[str]
    relationship2: Optional[str]
    address2: Optional[str]


class Sp(DurationModel):
    school_name: str
    study_level: str
    study_field: str
    province: str
    city: str
    address: str
    dli: str
    student_id: str
    tuition_cost: Optional[str]
    room_cost: Optional[str]
    other_cost: Optional[str]
    fund_available: str
    paid_person: str
    other: Optional[str]


class Wp(DurationModel):
    work_permit_type: str
    other_explain: Optional[str]
    employer_name: Optional[str]
    employer_address: Optional[str]
    work_province: Optional[str]
    work_city: Optional[str]
    work_address: Optional[str]
    job_title: Optional[str]
    brief_duties: Optional[str]
    lmia_num_or_offer_num: Optional[str]
    # pnp_certificated: bool
    caq_number: Optional[str]
    expiry_date: Optional[date]

    @root_validator
    def checkLMIA_OfferNum(cls, values):
        work_permit_type = values.get("work_permit_type")
        lmia_num_or_offer_num = values.get("lmia_num_or_offer_num")
        if (
            work_permit_type == "Labour Market Impact Assessment Stream"
            and not lmia_num_or_offer_num
        ):
            raise ValueError(
                "Since the work permit type is LMIA stream, LMIA number is required"
            )
        if (
            work_permit_type
            in [
                "Exemption from Labour Market Impact Assessment",
                "Live-in Caregiver Program",
                "Start-up Business Class",
            ]
            and not lmia_num_or_offer_num
        ):
            raise ValueError(
                "Since you are applying employer specific work permit, so the number of Employer portal offer of employment is required"
            )

        return values


class VrInCanada(DurationModel):
    application_purpose: str
    visit_purpose: str
    other_explain: Optional[str]
    funds_available: int
    paid_person: str
    other: Optional[str]
    name1: str
    relationship1: str
    address1: str
    name2: Optional[str]
    relationship2: Optional[str]
    address2: Optional[str]
    reason_for_vr: Optional[str]


class SpInCanada(DurationModel):
    application_purpose: str
    apply_work_permit: bool
    work_permit_type: Optional[str]
    caq_number: Optional[str]
    expiry_date: Optional[date]
    school_name: str
    study_level: str
    study_field: str
    province: str
    city: str
    address: str
    dli: str
    student_id: str
    tuition_cost: Optional[str]
    room_cost: Optional[str]
    other_cost: Optional[str]
    fund_available: str
    paid_person: str
    other: Optional[str]


class WpInCanada(DurationModel):
    application_purpose: str
    caq_number: Optional[str]
    expiry_date: Optional[date]
    work_permit_type: str
    employer_name: Optional[str]
    employer_address: Optional[str]
    work_province: Optional[str]
    work_city: Optional[str]
    work_address: Optional[str]
    job_title: Optional[str]
    brief_duties: Optional[str]
    lmia_num_or_offer_num: Optional[str]
    pnp_certificated: bool

    @root_validator
    def checkLMIA_OfferNum(cls, values):
        work_permit_type = values.get("work_permit_type")
        lmia_num_or_offer_num = values.get("lmia_num_or_offer_num")
        if (
            work_permit_type == "Labour Market Impact Assessment Stream"
            and not lmia_num_or_offer_num
        ):
            raise ValueError(
                "Since the work permit type is LMIA stream, LMIA number is required"
            )
        if (
            work_permit_type
            in [
                "Exemption from Labour Market Impact Assessment",
                "Live-in Caregiver Program",
                "Start-up Business Class",
            ]
            and not lmia_num_or_offer_num
        ):
            raise ValueError(
                "Since you are applying employer specific work permit, so the number of Employer portal offer of employment is required"
            )

        return values


class TrBackground(BaseModel):
    q1a: bool
    q1b: bool
    q1c: Optional[str]
    q2a: bool
    q2b: bool
    q2c: bool
    q2d: Optional[str]
    q3a: bool
    q3b: Optional[str]
    q4a: bool
    q4b: Optional[str]
    q5: bool
    q6: bool



from pydantic import BaseModel
from typing import Optional
from pydantic.class_validators import root_validator


class PositionBase(BaseModel):
    why_hire: str
    who_current_fill: str
    how_did_you_find: str
    how_when_offer: str
    worked_working: bool
    worked_working_details: Optional[str]
    has_same: bool
    lowest: Optional[float]
    highest: Optional[float]
    under_cba: bool
    which_union: Optional[str]
    lmia_refused: bool
    lmia_refused_reason: Optional[str]

    @root_validator
    def checkAnswers(cls, values):
        questions = ["worked_working", "under_cba", "lmia_refused"]
        explanations = ["worked_working_details", "which_union", "lmia_refused_reason"]
        qas = dict(zip(questions, explanations))
        for k, v in qas.items():
            if values.get(k) and not values.get(v):
                raise ValueError(
                    f"Since {k} is true, but you did not answer the question {v} in info-position sheet"
                )
        return values

    @root_validator
    def checkRange(cls, values):
        if values.get("has_same") == True:
            if not values["lowest"] or not values["highest"]:
                raise ValueError(
                    f"Since you indicated in info-position sheet has same employee in this position, you have to input lowest and highest wage."
                )
            elif values["lowest"] > values["highest"]:
                raise ValueError(
                    "range must have two number ,and the first is lower, the second is higher"
                )
            else:
                return values
        return values



# import context
import argparse, os
from termcolor import colored
from model.recruit.joboffer import JobofferModel, JobofferModelDocxAdapater
from model.recruit.jobad import JobadModel, JobadModelDocxAdapater
from model.recruit.recruitmentsummary import (
    RecruitmnetSummaryModel,
    RecruitmnetSummaryDocxAdaptor,
)


def creatExcel(model, excel_file):
    model(output_excel_file=excel_file)
    print(
        colored(
            f"{excel_file} is not existed, so we created the excel file based on your data structure.",
            "green",
        )
    )
    print(colored("Please fill the excel with data and do it again", "yellow"))


def generateDocx(source_excel, model, adapter, outpout, template_no):
    context = model(excels=[source_excel])
    context = adapter(context)
    context.make(outpout, template_no)


models = {
    "ja": {"model": JobadModel, "adaptor": JobadModelDocxAdapater},
    "jo": {"model": JobofferModel, "adaptor": JobofferModelDocxAdapater},
    "rs": {"model": RecruitmnetSummaryModel, "adaptor": RecruitmnetSummaryDocxAdaptor},
}


def main():
    """arguments:
    -e: excel file as source
    -t: output word file name
    -d: to generate docs--> ja: job advertisement docx, jo: job offer docx,rs: recruitment summary docx
    """
    parser = argparse.ArgumentParser(
        description="used for processing recruitment documents"
    )
    parser.add_argument(
        "-e",
        "--excel",
        help="input excel file name including the data for your specific stream",
    )
    parser.add_argument("-t", "--to", help="input docx file name for output")
    parser.add_argument("-tn", "--template_no", help="input template number (1,2)")
    parser.add_argument(
        "-d",
        "--document",
        help="input which kind of document to generate. type list: ja: job ad, jo: job offer, rs: recruitment summary",
    )
    args = parser.parse_args()

    # defautl template number is 1
    tn = args.template_no or 1

    if args.document and args.document not in models.keys():
        print(
            colored(
                f"{args.document} is not a valid document type. Valid type list is: ci, ert, eet, jd, jof, rep",
                "red",
            )
        )
        return

    # if input excel file, then generate it
    if args.excel and not os.path.isfile(args.excel):
        model = models[args.document]["model"]
        creatExcel(model, args.excel)
        print(colored(f"{args.excel} is created...", "green"))
        return

    if args.excel and args.to and args.document:
        model = models[args.document]["model"]
        adaptor = models[args.document]["adaptor"]
        generateDocx(args.excel, model, adaptor, args.to, template_no=tn)
        return
    print(colored("Wrong command/flag combination, type recruit -h for help", "red"))


if __name__ == "__main__":
    main()



from pydantic import validator, EmailStr
from typing import Optional, List
from model.common.employerbase import EmployerBase
from model.common.jobofferbase import JobofferBase
from model.common.address import Address, Addresses
from model.common.utils import makeList


class General(EmployerBase):
    company_intro: str
    business_intro: str
    recruit_email: EmailStr


class JobOffer(JobofferBase):
    # education_level:Optional[str]
    disability_insurance: bool
    dental_insurance: bool
    empolyer_provided_persion: bool
    extended_medical_insurance: bool
    extra_benefits: Optional[str]
    duties_brief: str
    duties: list
    specific_edu_requirement: str
    skill_experience_requirement: str
    other_requirements: Optional[list]

    _str2bool_duties = validator("duties", allow_reuse=True, pre=True)(makeList)
    _str2bool_other_requirements = validator(
        "other_requirements", allow_reuse=True, pre=True
    )(makeList)

    @property
    def benefits(self):
        bs = []
        if self.disability_insurance:
            bs.append("Disability insurance")
        if self.dental_insurance:
            bs.append("Dental insurance")
        if self.empolyer_provided_persion:
            bs.append("Employer provided persion")
        if self.extended_medical_insurance:
            bs.append("Extended medical insurance")
        if self.extra_benefits:
            bs.append(self.extra_benefits)
        return ", ".join(bs)


class ErAddress(Address):
    def __init_subclass__(cls) -> None:
        return super().__init_subclass__()


class ErAddresses(Addresses):
    def __init__(self, address_list: List[Address]) -> None:
        super().__init__(address_list)



from typing import List, Optional, Union
from model.common.address import Address
from model.common.educationbase import EducationBase
from model.common.phone import Phone
from model.common.commonmodel import CommonModel
from model.common.employmentbase import EmploymentBase
from model.common.jobofferbase import JobofferBase
from model.common.id import ID
from model.common.person import Person
from model.common.language import LanguageBase
from datetime import date, datetime
from pydantic import BaseModel, EmailStr, root_validator, validator
from model.common.utils import normalize
from webform.prportal.data.country import country_residence
from termcolor import colored


class PersonId(ID):
    pass


class Rcic(BaseModel):
    first_name: str
    last_name: str
    telephone: str


class Language(LanguageBase):
    date: Optional[date]
    report_date: Optional[date]
    registration_number: Optional[str]
    pin: Optional[str]


class General(BaseModel):
    legal_name: str
    operating_name: Optional[str]


class ErAddress(Address):
    pass


class Joboffer(JobofferBase):
    phone: str
    is_working: bool


class Employment(EmploymentBase):
    bcpnp_qualified: bool


class Marriage(BaseModel):
    marital_status: Optional[str]
    married_date: Optional[date]
    sp_last_name: Optional[str]
    sp_first_name: Optional[str]
    sp_in_canada: Optional[bool]
    sp_canada_status: Optional[str]
    sp_canada_status_end_date: Optional[date]
    sp_in_canada_other: Optional[str]
    sp_in_canada_work: Optional[bool]
    sp_canada_occupation: Optional[str]
    sp_canada_employer: Optional[str]

    @root_validator
    def checkCurrentMarriage(cls, values):
        marital_status = values["marital_status"]
        married_date = values["married_date"]
        sp_last_name = values["sp_last_name"]
        sp_first_name = values["sp_first_name"]

        if marital_status in ["Common-Law", "Married"] and not all(
            [married_date, sp_last_name, sp_first_name]
        ):
            raise ValueError(
                f"Marital stauts is {marital_status}, but you did not answer all the questions for spouse: first/last name, or marriage date"
            )
        return values

    @root_validator
    def checkCanadaStatus(cls, values):
        sp_in_canada = values["sp_in_canada"]
        sp_canada_status = values["sp_canada_status"]
        sp_canada_status_end_date = values["sp_canada_status_end_date"]
        sp_in_canada_other = values["sp_in_canada_other"]

        if sp_in_canada and not all([sp_canada_status, sp_canada_status_end_date]):
            raise ValueError(
                f"Spouse is in Canada, but you did not answer all the questions for spouse status and/or status end date",
                "red",
            )
        if sp_canada_status == "Other" and not sp_in_canada_other:
            raise ValueError(
                f"Spouse is in Canada and status is other, but you did not give the explanation"
            )
        return values

    @root_validator
    def checkWorkInCanada(cls, values):
        sp_in_canada_work = values["sp_in_canada_work"]
        sp_canada_occupation = values["sp_canada_occupation"]
        sp_canada_employer = values["sp_canada_employer"]

        if sp_in_canada_work and not all([sp_canada_occupation, sp_canada_employer]):
            raise ValueError(
                colored(
                    f"Spouse is working in Canada, but you did not answer all the questions for spouse occupation and/or employer",
                    "red",
                )
            )
        return values


class Personal(Person):
    email: EmailStr
    used_last_name: Optional[str]
    used_first_name: Optional[str]
    place_of_birth: str
    country_of_birth: str
    did_eca: bool
    eca_supplier: Optional[str]
    eca_number: Optional[str]
    ita_assessed: bool
    ita_assess_number: Optional[str]

    _normalize_used_first_name = validator(
        "used_first_name", allow_reuse=True, check_fields=False
    )(normalize)
    _normalize_used_last_name = validator(
        "used_last_name", allow_reuse=True, check_fields=False
    )(normalize)

    @root_validator
    def checkECA(cls, values):
        did_eca = values.get("did_eca")
        eca_supplier = values.get("eca_supplier")
        eca_number = values.get("eca_number")

        if did_eca and not all([eca_supplier, eca_number]):
            raise ValueError(f"Did ECA,but not input ECA number and/or ECA supplier")
        return values

    @root_validator
    def checkITA(cls, values):
        ita_assessed = values.get("ita_assessed")
        ita_assess_number = values.get("ita_assess_number")

        if ita_assessed and not ita_assess_number:
            raise ValueError(f"You selected ITA as Yes,but not input ITA assessment number")
        return values

    @property
    def user_id(self):
        dob = self.dob.strftime(("%Y-%m-%d"))
        return (
            self.last_name[0].upper()
            + self.first_name[0]
            + dob.split("-")[0]
            + dob.split("-")[1]
            + dob.split("-")[2]
        )

    @property
    def password(self):
        return "Super" + str(datetime.today().year) + "!"


class Education(EducationBase):
    city: Optional[str]
    country: Optional[str]
    is_trade: Optional[bool]


class Bcpnp(BaseModel):
    account: Optional[str]
    password: Optional[str]
    # has_current_app: bool
    has_applied_before: bool
    pre_file_no: Optional[str]
    submission_date: Optional[date]
    case_stream: Optional[str]
    q1: bool
    q1_explaination: Optional[str]
    q2: bool
    q2_explaination: Optional[str]
    q3: bool
    q3_explaination: Optional[str]
    q4: bool
    q4_file_number: Optional[str]
    q4_explaination: Optional[str]
    q5: bool
    q5_explaination: Optional[str]
    q6: bool
    q6_explaination: Optional[str]
    q7: bool
    q7_explaination: Optional[str]


class Ee(BaseModel):
    ee_profile_no: str
    ee_expiry_date: date
    ee_jsvc: str
    ee_score: str
    ee_noc: str
    ee_job_title: str


class BcpnpModelReg(CommonModel):
    personal: Personal
    general: General
    joboffer: Joboffer
    address: List[Address]
    eraddress: List[ErAddress]
    phone: List[Phone]
    personid: List[PersonId]
    bcpnp: Bcpnp
    education: List[Education]
    employment: List[Employment]
    language: List[Language]
    rcic: Rcic
    marriage: Marriage

    # initialize the model with a list of excels, which includes all nececcery information the model required. if outpuot_excel_file is not None, it will make an excel file.
    def __init__(self, excels=None, output_excel_file=None):
        if output_excel_file:
            excels = self.getExcels(
                ["excel/er.xlsx", "excel/pa.xlsx", "excel/bcpnp.xlsx", "excel/rep.xlsx"]
            )
        else:
            if excels is None or len(excels) == 0:
                raise ValueError(
                    "You must input excel file list as source data for validation"
                )
        super().__init__(excels, output_excel_file, globals())


class BcpnpEEModelReg(BcpnpModelReg):
    ee: Ee

    def __init__(self, excels=None, output_excel_file=None):
        super().__init__(excels=excels, output_excel_file=output_excel_file)



import json, copy
from termcolor import colored
from openpyxl import load_workbook, Workbook
from openpyxl.workbook.views import BookView
from openpyxl.utils import get_column_letter
from collections import OrderedDict
from source.infosheet import SheetDict
from source.tablesheet import TableList

# from source.style import style
# from source.format import Format
from source.excelwriter import ExcelWritter


class Excel(object):
    """Excel manupulate excel for getting source data, compare excels or sheets data, and write new excel for collecting data.

    Args:
        excel_name ([str]): [an excel file name]
        sheets ([list]): [default is None, meaning will get all sheets; If assing a list of sheet names, it will only access the specified sheets.
        sheet name in this Excel class will with its original 'info-' or 'table-' head, but the head will be removed when save to dict for exporting;
        When write to a new excel file, the sheet name of a object of SheetDict will be insert 'info-' in the name, while for TableList object, 'table-'
        will be inserted too]
    Returns: an Excel object
    """

    def __init__(self, excel_name, sheets=None):
        self.excel_name = excel_name
        self.wb = load_workbook(self.excel_name)
        self.sheetNames = (
            self.wb.sheetnames if sheets == None else sheets
        )  # if not specify sheets, it will handle all sheets
        self.sheet_names = [
            sheetName
            for sheetName in self.sheetNames
            if sheetName.lower().startswith("info-")
        ]  # save info sheet names
        self.table_names = [
            sheetName
            for sheetName in self.sheetNames
            if sheetName.lower().startswith("table-")
        ]  # save table sheet names
        self.sheets = OrderedDict()  # save info sheet objects
        self.tables = OrderedDict()  # save table sheet objects
        self._getSheetsTables()  # get all info and table sheets objects and save to sheets and tables

    def __str__(self):
        return (
            self.excel_name
            + " with info sheets: "
            + ", ".join(self.sheets)
            + " , and table sheets: "
            + " ,".join(self.tables)
        )

    # = operator. Return True if self and another excel have same sheets, and every sheet has same variables
    def __eq__(self, another: object):
        if self.sheetNames != another.sheetNames:
            return False  # if not all sheet names same, False
        # check if info sheets are same
        pairs = list(zip(self.sheets, another.sheets))
        for self_pair, another_pair in pairs:
            if self_pair != another.pair:
                return False
        # check if table sheets are same
        pairs = list(zip(self.tables, another.tables))
        for self_pair, another_pair in pairs:
            if self_pair != another.pair:
                return False
        return True

    # + union two excel objs (self | another), and return a new excel obj
    # in excel level, I consider to merge value together with variables. If duplicated, another value will prevail.
    def __add__(self, another: object):
        # iterate another and insert or replace self's
        # merge sheets (info and talbe)
        for sheet in another.sheets:
            if sheet not in self.sheets:  # sheet is a key of sheets OrderedDict
                self.sheets[sheet] = another.sheets[sheet]
            else:
                self.sheets[sheet] += another.sheets[sheet]
        for table in another.tables:
            if table not in self.tables:
                # 创建一个新的table
                self.tables[table] = another.tables[table]
            else:
                self.tables[table] += another.tables[
                    table
                ]  # 原来是每个子项都自己算，v1.0.3不做了。 从V1.0.4改回来。
                # append rows to old one; 直接把后面的excel table里面的rows添加到签前面的table里面。
                # self.tables[table].data.extend(another.tables[table].data) # 这个是添加到后面的版本。 只有在V1.0.3里面用。

        self.sheet_names += another.sheet_names
        self.table_names += another.table_names
        return self

    # - set difference, return a new obj with sheets in self but not in another, and with variables in self sheet but not in another if has same sheet
    def __sub__(self, another: object):
        # for sheet in obj.sheets:
        for sheet in self.sheets:
            if (
                sheet in another.sheets
            ):  # if sheet in obj and in another, self sheet minus another sheet. sheet is a key of sheets OrderedDict
                self.sheets[sheet] -= another.sheets[sheet]
            # if len(self.sheets[sheet])==0:   # if a sheet is empty, delete it
            #     self.sheets.pop(sheet)

        # table part: for columns sub, only care about variables; while for rows, care about if same row
        # for table in self.tables:
        for table in self.tables:
            if (
                table in another.tables
            ):  # if table in obj and in another, self sheet minus another sheet. sheet is a key of sheets OrderedDict
                self.tables[table] -= another.tables[table]
            # if len(self.tables[table])==0:
            #     self.tables.pop(table)
        self.sheet_names = list(set(self.sheet_names) - set(another.sheet_names))
        self.table_names = list(set(self.table_names) - set(another.table_names))

        return self

    # copy another's excel's common contents to self
    def copy(self, another: object):
        # iterate another's sheet, replace self's value if it is in self too
        for sheet in another.sheets:
            if sheet in self.sheets:  # sheet is a key of sheets OrderedDict
                self.sheets[sheet] = self.sheets[sheet].copy(another.sheets[sheet])
        for table in another.tables:
            if table in self.tables:
                self.tables[table] = self.tables[table].copy(another.tables[table])
        return self

    """
    Below are methods for getting data from excel files and export 
    """
    # get sheet variable name from sheet name
    def _getVariable(self, talbe_sheet_name):
        if talbe_sheet_name.startswith("info-"):
            return talbe_sheet_name[5:]
        elif talbe_sheet_name.startswith("table-"):
            return talbe_sheet_name[6:]
        else:
            return talbe_sheet_name

    # check if a row is None. Return True if there is at least one column not None
    def _notNone(self, row_data):
        return any([False if x == None else True for x in row_data])

    # get info sheet object
    def _getSheet(self, sheet):
        values = list(self.wb[sheet].values)
        values = [x for x in values if self._notNone(x)]  # clean data by removing None
        sheet_obj = SheetDict(values)
        setattr(
            sheet_obj, "name", self._getVariable(sheet)
        )  # insert a sheet's name by removing 'info-'
        return sheet_obj

    # get table sheet object
    def _getTable(self, table):
        values = list(self.wb[table].values)
        values = [x for x in values if self._notNone(x)]  # clean data by removing None
        table_obj = TableList(values)
        setattr(
            table_obj, "name", self._getVariable(table)
        )  # insert a table's name by removing 'table-'
        return table_obj

    # get the excel's specified sheets(info and table). In this step, the 'info-' and 'table-' will be removed. The rest sheet name
    # will be used as the key of the object of SheetDict or TableList.
    def _getSheetsTables(self):
        for sn in self.sheet_names:
            self.sheets[self._getVariable(sn)] = self._getSheet(sn)
        for tn in self.table_names:
            self.tables[self._getVariable(tn)] = self._getTable(tn)

    # get sheet object from the dict of self.sheets by specific sheet name,and variable list. default None means all variables. This name is variable name, without info-
    def getSheet(self, sheet_name, variables=None):
        sheet = self.sheets.get(sheet_name)
        if sheet:
            if variables:
                data = {k: v for k, v in sheet.data.items() if k in variables}
                sheet.data = data
                return sheet
            else:
                return sheet
        elif sheet == None:
            return None
        elif len(sheet) == 0:
            return sheet
        else:
            raise ValueError(
                f"Sheet info-{sheet_name} is not existed in {self.excel_name}. Or it was not referrenced in your pydantic models"
            )

    # get sheet object by specific sheet name. This name is variable name, without table-
    def getTable(self, table_name, variables=None):
        table = self.tables.get(table_name)
        if table is not None:
            if variables:
                temp_table = copy.deepcopy(table)
                if len(table.data) > 0:
                    for index, data in enumerate(table.data):
                        for k in data.__dict__:  # check every TableNode object
                            if k not in variables:  # if not in variable list
                                delattr(
                                    temp_table.data[index], k
                                )  # delete from returning obj
                temp_table.column_variables = variables  # update the table variables
                temp_table.column_titles = OrderedDict(
                    {k: v for k, v in table.column_titles.items() if k in variables}
                )  # update the table titles
                return temp_table
            else:
                return table
        else:
            raise ValueError(
                f"Table table-{table_name} is not existed in {self.excel_name}. Or it was not referrenced in your pydantic models"
            )

    # get sheet dict by specific sheet name. This name is variable name, without info-
    def getSheetData(self, sheet_name):
        # sheet_data=OrderedDict()
        if self.sheets.get(sheet_name):
            sheet = self.sheets.get(sheet_name)
            # get key(variable) and value from object
            sheet_data = {k: v.value for k, v in sheet.data.items()}
        elif len(self.sheets.get(sheet_name).data) == 0:
            sheet_data = {}
        else:
            raise ValueError(
                f"Sheet info-{sheet_name} is not existed in {self.excel_name}. Or it was not referrenced in your pydantic models"
            )

        return sheet_data

    # get table list by specific sheet name. This name is variable name, without table-
    def getTableData(self, table_name):
        table = self.tables.get(table_name, None)
        # else:
        #     raise ValueError(f"Table table-{table_name} is not existed in {self.excel_name}. Or it was not referrenced in your pydantic models")
        # if table is not existed or data is [], will return []
        if table:
            table_data = [td.__dict__ for td in table.data]
            return table_data
        return []

    # get all info-sheets and table-sheets data
    @property
    def dict(self):
        the_data = {
            **{sheet: self.getSheetData(sheet) for sheet in self.sheets.keys()},
            **{table: self.getTableData(table) for table in self.tables.keys()},
        }
        return the_data

    @property
    def json(self):
        print(json.dumps(self.dict, indent=3, default=str))

    """
    Below is write back to excel  mothods
    """

    def cellStyle(self, cell, format):
        cell.font = format.get("font")
        cell.border = format.get("border")
        cell.fill = format.get("fill")
        cell.alignment = format.get("alignment")

    def _formatCells(self, ws, columns):
        for col in range(1, columns + 1):
            cell = ws.cell(1, col)
            self.cellStyle(cell, style.get("TITLE_FORMAT"))

    def makeInfoSheet(self, sheet: SheetDict):
        # create a new info sheet in new workbook
        sheet_name = "info-" + sheet.name
        new_ws = self.new_wb.create_sheet(sheet_name)
        rows = []
        rows.append([sheet.sheet_title])
        rows.append(list(sheet.column_titles))
        # not use titled variables
        # titled_variables=[ x.title() for x in sheet.column_variables]
        rows.append(sheet.column_variables)
        for new_row in sheet.data.values():
            rows.append(list(new_row.__dict__.values()))
        for row in rows:
            new_ws.append(row)
        max_column = get_column_letter(len(sheet.column_titles))
        new_ws.merge_cells(f"A1:{max_column}1")

    def makeTableSheet(self, table: TableList):
        # create a new table sheet in new workbook
        table_name = "table-" + table.name
        new_ws = self.new_wb.create_sheet(table_name)
        rows = []
        rows.append([table.sheet_title])
        variables = table.column_variables
        titles = len(variables) > 0 and [table.column_titles.get(v) for v in variables]
        rows.append(titles)
        rows.append(variables)
        for new_row in table.data:
            rows.append(list(new_row.__dict__.values()))
        for row in rows:
            row and new_ws.append(row)
        max_column = get_column_letter(len(table.column_titles))
        new_ws.merge_cells(f"A1:{max_column}1")

    def makeExcel(self, file_name, protection=False, sheets=None, tables=None):
        # 创建新的excel文件
        self.new_wb = Workbook()

        self.new_wb.views = [
            BookView(xWindow=0, yWindow=0, windowWidth=18140, windowHeight=15540)
        ]

        # handle info sheets. if None, including all sheets/tables, if [], including nothing. if has sheets/tables, including those sheets/tables
        sheets = sheets if sheets is not None else self.sheets.values()
        tables = tables if tables is not None else self.tables.values()
        [self.makeInfoSheet(sheet) for sheet in sheets if sheet is not None]
        # handle table sheets
        [self.makeTableSheet(table) for table in tables if table is not None]
        # 保存为新文件
        if len(self.sheets) > 0 or len(self.tables) > 0:
            self.new_wb.remove(
                self.new_wb["Sheet"]
            )  # exists only when niether info nor table has sheet
        # old version, removed
        # f=Format(self.new_wb)
        # f.setAll()
        # if protection:
        #     f.protectAll()
        self.new_wb.save(file_name)  # 先保存
        self.formatExcel(file_name, protection)  # format这个文件

    def formatExcel(self, output_file_name, protection=True):
        input_excel_obj = Excel(output_file_name)  # get excel obj for format
        ew = ExcelWritter(input_excel_obj, output_file_name)
        if protection:
            password = input("please input password:")
            ew.create(protection, password)
        else:
            ew.create(protection=False)
